from datetime import timedelta

import django_filters
from django import forms
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _

from wagtail.admin.auth import user_has_any_page_permission, user_passes_test
from wagtail.admin.filters import DateRangePickerWidget, WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.log_actions import registry as log_action_registry
from wagtail.models import (
    Page,
    PageLogEntry,
    PageRevision,
    TaskState,
    UserPagePermissionsProxy,
    WorkflowState,
)


class PageHistoryReportFilterSet(WagtailFilterSet):
    action = django_filters.ChoiceFilter(choices=log_action_registry.get_choices)
    hide_commenting_actions = django_filters.BooleanFilter(
        label=_("Hide commenting actions"),
        method="filter_hide_commenting_actions",
        widget=forms.CheckboxInput,
    )
    user = django_filters.ModelChoiceFilter(
        field_name="user",
        queryset=lambda request: PageLogEntry.objects.all().get_users(),
    )
    timestamp = django_filters.DateFromToRangeFilter(
        label=_("Date"), widget=DateRangePickerWidget
    )

    def filter_hide_commenting_actions(self, queryset, name, value):
        if value:
            queryset = queryset.exclude(action__startswith="wagtail.comments")
        return queryset

    class Meta:
        model = PageLogEntry
        fields = ["action", "user", "timestamp", "hide_commenting_actions"]


def workflow_history(request, page_id):
    page = get_object_or_404(Page, id=page_id)

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_edit():
        raise PermissionDenied

    workflow_states = WorkflowState.objects.filter(page=page).order_by("-created_at")

    paginator = Paginator(workflow_states, per_page=20)
    workflow_states = paginator.get_page(request.GET.get("p"))

    return TemplateResponse(
        request,
        "wagtailadmin/pages/workflow_history/index.html",
        {
            "page": page,
            "workflow_states": workflow_states,
        },
    )


def workflow_history_detail(request, page_id, workflow_state_id):
    page = get_object_or_404(Page, id=page_id)

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_edit():
        raise PermissionDenied

    workflow_state = get_object_or_404(WorkflowState, page=page, id=workflow_state_id)

    # Get QuerySet of all revisions that have existed during this workflow state
    # It's possible that the page is edited while the workflow is running, so some
    # tasks may be repeated. All tasks that have been completed no matter what
    # revision needs to be displayed on this page.
    page_revisions = PageRevision.objects.filter(
        page=page,
        id__in=TaskState.objects.filter(workflow_state=workflow_state).values_list(
            "page_revision_id", flat=True
        ),
    ).order_by("-created_at")

    # Now get QuerySet of tasks completed for each revision
    task_states_by_revision_task = [
        (
            page_revision,
            {
                task_state.task: task_state
                for task_state in TaskState.objects.filter(
                    workflow_state=workflow_state, page_revision=page_revision
                )
            },
        )
        for page_revision in page_revisions
    ]

    # Make sure task states are always in a consistent order
    # In some cases, they can be completed in a different order to what they are defined
    tasks = workflow_state.workflow.tasks.all()
    task_states_by_revision = [
        (page_revision, [task_states_by_task.get(task, None) for task in tasks])
        for page_revision, task_states_by_task in task_states_by_revision_task
    ]

    # Generate timeline
    completed_task_states = (
        TaskState.objects.filter(workflow_state=workflow_state)
        .exclude(finished_at__isnull=True)
        .exclude(status=TaskState.STATUS_CANCELLED)
    )

    timeline = [
        {
            "time": workflow_state.created_at,
            "action": "workflow_started",
            "workflow_state": workflow_state,
        }
    ]

    if workflow_state.status not in (
        WorkflowState.STATUS_IN_PROGRESS,
        WorkflowState.STATUS_NEEDS_CHANGES,
    ):
        last_task = completed_task_states.order_by("finished_at").last()
        if last_task:
            timeline.append(
                {
                    "time": last_task.finished_at + timedelta(milliseconds=1),
                    "action": "workflow_completed",
                    "workflow_state": workflow_state,
                }
            )

    for page_revision in page_revisions:
        timeline.append(
            {
                "time": page_revision.created_at,
                "action": "page_edited",
                "revision": page_revision,
            }
        )

    for task_state in completed_task_states:
        timeline.append(
            {
                "time": task_state.finished_at,
                "action": "task_completed",
                "task_state": task_state,
            }
        )

    timeline.sort(key=lambda t: t["time"])
    timeline.reverse()

    return TemplateResponse(
        request,
        "wagtailadmin/pages/workflow_history/detail.html",
        {
            "page": page,
            "workflow_state": workflow_state,
            "tasks": tasks,
            "task_states_by_revision": task_states_by_revision,
            "timeline": timeline,
        },
    )


class PageHistoryView(ReportView):
    template_name = "wagtailadmin/pages/history.html"
    title = _("Page history")
    header_icon = "history"
    paginate_by = 20
    filterset_class = PageHistoryReportFilterSet

    @method_decorator(user_passes_test(user_has_any_page_permission))
    def dispatch(self, request, *args, **kwargs):
        self.page = get_object_or_404(Page, id=kwargs.pop("page_id")).specific

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, object_list=None, **kwargs):
        context = super().get_context_data(*args, object_list=object_list, **kwargs)
        context["page"] = self.page
        context["subtitle"] = self.page.get_admin_display_title()
        context["page_latest_revision"] = self.page.get_latest_revision()

        return context

    def get_queryset(self):
        return PageLogEntry.objects.filter(page=self.page).select_related(
            "revision", "user", "user__wagtail_userprofile"
        )



# Generated by Django 3.1.2 on 2020-10-15 00:52

from django.db import migrations


def add_choose_permission_to_admin_groups(apps, _schema_editor):
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    Group = apps.get_model("auth.Group")

    # Get document content type
    document_content_type, _created = ContentType.objects.get_or_create(
        model="document", app_label="wagtaildocs"
    )

    # Create the Choose permission (if it doesn't already exist)
    choose_document_permission, _created = Permission.objects.get_or_create(
        content_type=document_content_type,
        codename="choose_document",
        defaults={"name": "Can choose document"},
    )

    # Assign it to all groups which have "Access the Wagtail admin" permission.
    # This emulates the previous behavior, where everyone who would access the admin
    # could choose any document in any Collection, because choosing wasn't permissioned.
    for group in Group.objects.filter(permissions__codename="access_admin"):
        group.permissions.add(choose_document_permission)


def remove_choose_permission(apps, _schema_editor):
    """Reverse the above additions of permissions."""
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    document_content_type = ContentType.objects.get(
        model="document",
        app_label="wagtaildocs",
    )
    # This cascades to Group
    Permission.objects.filter(
        content_type=document_content_type, codename="choose_document"
    ).delete()


def get_choose_permission(apps):
    Permission = apps.get_model("auth.Permission")
    ContentType = apps.get_model("contenttypes.ContentType")

    document_content_type, _created = ContentType.objects.get_or_create(
        model="document",
        app_label="wagtaildocs",
    )
    return Permission.objects.filter(
        content_type=document_content_type, codename__in=["choose_document"]
    ).first()


def copy_choose_permission_to_collections(apps, _schema_editor):
    Collection = apps.get_model("wagtailcore.Collection")
    Group = apps.get_model("auth.Group")
    GroupCollectionPermission = apps.get_model("wagtailcore.GroupCollectionPermission")

    root_collection = Collection.objects.get(depth=1)

    permission = get_choose_permission(apps)
    if permission:
        for group in Group.objects.filter(permissions=permission):
            GroupCollectionPermission.objects.create(
                group=group, collection=root_collection, permission=permission
            )


def remove_choose_permission_from_collections(apps, _schema_editor):
    GroupCollectionPermission = apps.get_model("wagtailcore.GroupCollectionPermission")
    choose_permission = get_choose_permission(apps)
    if choose_permission:
        GroupCollectionPermission.objects.filter(permission=choose_permission).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("wagtaildocs", "0010_document_file_hash"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="document",
            options={
                "permissions": [("choose_document", "Can choose document")],
                "verbose_name": "document",
                "verbose_name_plural": "documents",
            },
        ),
        migrations.RunPython(
            add_choose_permission_to_admin_groups, remove_choose_permission
        ),
        migrations.RunPython(
            copy_choose_permission_to_collections,
            remove_choose_permission_from_collections,
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations


def add_image_permissions_to_admin_groups(apps, schema_editor):
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    Group = apps.get_model("auth.Group")

    # Get image permissions
    image_content_type, _created = ContentType.objects.get_or_create(
        model="image", app_label="wagtailimages"
    )

    add_image_permission, _created = Permission.objects.get_or_create(
        content_type=image_content_type,
        codename="add_image",
        defaults={"name": "Can add image"},
    )
    change_image_permission, _created = Permission.objects.get_or_create(
        content_type=image_content_type,
        codename="change_image",
        defaults={"name": "Can change image"},
    )
    delete_image_permission, _created = Permission.objects.get_or_create(
        content_type=image_content_type,
        codename="delete_image",
        defaults={"name": "Can delete image"},
    )

    # Assign it to Editors and Moderators groups
    for group in Group.objects.filter(name__in=["Editors", "Moderators"]):
        group.permissions.add(
            add_image_permission, change_image_permission, delete_image_permission
        )


def remove_image_permissions(apps, schema_editor):
    """Reverse the above additions of permissions."""
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    image_content_type = ContentType.objects.get(
        model="image",
        app_label="wagtailimages",
    )
    # This cascades to Group
    Permission.objects.filter(
        content_type=image_content_type,
        codename__in=("add_image", "change_image", "delete_image"),
    ).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0001_initial"),
        # Need to run wagtailcores initial data migration to make sure the groups are created
        ("wagtailcore", "0002_initial_data"),
    ]

    operations = [
        migrations.RunPython(
            add_image_permissions_to_admin_groups, remove_image_permissions
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-11 12:03
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0013_make_rendition_upload_callable"),
    ]

    operations = [
        migrations.AddField(
            model_name="rendition",
            name="filter_spec",
            field=models.CharField(
                blank=True, db_index=True, default="", max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="rendition",
            name="filter",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="wagtailimages.Filter",
            ),
        ),
    ]



import itertools
import uuid
from collections import OrderedDict, defaultdict
from collections.abc import MutableSequence

from django import forms
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.forms.utils import ErrorList
from django.utils.functional import cached_property
from django.utils.html import format_html_join
from django.utils.translation import gettext as _

from wagtail.admin.staticfiles import versioned_static
from wagtail.telepath import Adapter, register

from .base import Block, BoundBlock, DeclarativeSubBlocksMetaclass, get_help_icon

__all__ = [
    "BaseStreamBlock",
    "StreamBlock",
    "StreamValue",
    "StreamBlockValidationError",
]


class StreamBlockValidationError(ValidationError):
    def __init__(self, block_errors=None, non_block_errors=None):
        self.non_block_errors = non_block_errors
        self.block_errors = block_errors

        params = {}
        if block_errors:
            params.update(block_errors)
        if non_block_errors:
            params[NON_FIELD_ERRORS] = non_block_errors
        super().__init__("Validation error in StreamBlock", params=params)


class StreamBlockValidationErrorAdapter(Adapter):
    js_constructor = "wagtail.blocks.StreamBlockValidationError"

    def js_args(self, error):
        return [
            error.non_block_errors.as_data(),
            {
                block_id: child_errors.as_data()
                for block_id, child_errors in error.block_errors.items()
            },
        ]

    @cached_property
    def media(self):
        return forms.Media(
            js=[
                versioned_static("wagtailadmin/js/telepath/blocks.js"),
            ]
        )


register(StreamBlockValidationErrorAdapter(), StreamBlockValidationError)


class BaseStreamBlock(Block):
    def __init__(self, local_blocks=None, **kwargs):
        self._constructor_kwargs = kwargs

        super().__init__(**kwargs)

        # create a local (shallow) copy of base_blocks so that it can be supplemented by local_blocks
        self.child_blocks = self.base_blocks.copy()
        if local_blocks:
            for name, block in local_blocks:
                block.set_name(name)
                self.child_blocks[name] = block

    def get_default(self):
        """
        Default values set on a StreamBlock should be a list of (type_name, value) tuples -
        we can't use StreamValue directly, because that would require a reference back to
        the StreamBlock that hasn't been built yet.

        For consistency, then, we need to convert it to a StreamValue here for StreamBlock
        to work with.
        """
        return StreamValue(self, self.meta.default)

    def sorted_child_blocks(self):
        """Child blocks, sorted in to their groups."""
        return sorted(
            self.child_blocks.values(), key=lambda child_block: child_block.meta.group
        )

    def grouped_child_blocks(self):
        """
        The available child block types of this stream block, organised into groups according to
        their meta.group attribute.
        Returned as an iterable of (group_name, list_of_blocks) tuples
        """
        return itertools.groupby(
            self.sorted_child_blocks(), key=lambda child_block: child_block.meta.group
        )

    def value_from_datadict(self, data, files, prefix):
        count = int(data["%s-count" % prefix])
        values_with_indexes = []
        for i in range(0, count):
            if data["%s-%d-deleted" % (prefix, i)]:
                continue
            block_type_name = data["%s-%d-type" % (prefix, i)]
            try:
                child_block = self.child_blocks[block_type_name]
            except KeyError:
                continue

            values_with_indexes.append(
                (
                    int(data["%s-%d-order" % (prefix, i)]),
                    block_type_name,
                    child_block.value_from_datadict(
                        data, files, "%s-%d-value" % (prefix, i)
                    ),
                    data.get("%s-%d-id" % (prefix, i)),
                )
            )

        values_with_indexes.sort()
        return StreamValue(
            self,
            [
                (child_block_type_name, value, block_id)
                for (
                    index,
                    child_block_type_name,
                    value,
                    block_id,
                ) in values_with_indexes
            ],
        )

    def value_omitted_from_data(self, data, files, prefix):
        return ("%s-count" % prefix) not in data

    @property
    def required(self):
        return self.meta.required

    def clean(self, value):
        cleaned_data = []
        errors = {}
        non_block_errors = ErrorList()
        for i, child in enumerate(value):  # child is a StreamChild instance
            try:
                cleaned_data.append(
                    (child.block.name, child.block.clean(child.value), child.id)
                )
            except ValidationError as e:
                errors[i] = ErrorList([e])

        if self.meta.min_num is not None and self.meta.min_num > len(value):
            non_block_errors.append(
                ValidationError(
                    _("The minimum number of items is %d") % self.meta.min_num
                )
            )
        elif self.required and len(value) == 0:
            non_block_errors.append(ValidationError(_("This field is required.")))

        if self.meta.max_num is not None and self.meta.max_num < len(value):
            non_block_errors.append(
                ValidationError(
                    _("The maximum number of items is %d") % self.meta.max_num
                )
            )

        if self.meta.block_counts:
            block_counts = defaultdict(int)
            for item in value:
                block_counts[item.block_type] += 1

            for block_name, min_max in self.meta.block_counts.items():
                block = self.child_blocks[block_name]
                max_num = min_max.get("max_num", None)
                min_num = min_max.get("min_num", None)
                block_count = block_counts[block_name]
                if min_num is not None and min_num > block_count:
                    non_block_errors.append(
                        ValidationError(
                            "{}: {}".format(
                                block.label,
                                _("The minimum number of items is %d") % min_num,
                            )
                        )
                    )
                if max_num is not None and max_num < block_count:
                    non_block_errors.append(
                        ValidationError(
                            "{}: {}".format(
                                block.label,
                                _("The maximum number of items is %d") % max_num,
                            )
                        )
                    )

        if errors or non_block_errors:
            # The message here is arbitrary - outputting error messages is delegated to the child blocks,
            # which only involves the 'params' list
            raise StreamBlockValidationError(
                block_errors=errors, non_block_errors=non_block_errors
            )

        return StreamValue(self, cleaned_data)

    def to_python(self, value):
        # the incoming JSONish representation is a list of dicts, each with a 'type' and 'value' field
        # (and possibly an 'id' too).
        # This is passed to StreamValue to be expanded lazily - but first we reject any unrecognised
        # block types from the list
        return StreamValue(
            self,
            [
                child_data
                for child_data in value
                if child_data["type"] in self.child_blocks
            ],
            is_lazy=True,
        )

    def bulk_to_python(self, values):
        # 'values' is a list of streams, each stream being a list of dicts with 'type', 'value' and
        # optionally 'id'.
        # We will iterate over these streams, constructing:
        # 1) a set of per-child-block lists ('child_inputs'), to be sent to each child block's
        #    bulk_to_python method in turn (giving us 'child_outputs')
        # 2) a 'block map' of each stream, telling us the type and id of each block and the index we
        #    need to look up in the corresponding child_outputs list to obtain its final value

        child_inputs = defaultdict(list)
        block_maps = []

        for stream in values:
            block_map = []
            for block_dict in stream:
                block_type = block_dict["type"]

                if block_type not in self.child_blocks:
                    # skip any blocks with an unrecognised type
                    continue

                child_input_list = child_inputs[block_type]
                child_index = len(child_input_list)
                child_input_list.append(block_dict["value"])
                block_map.append((block_type, block_dict.get("id"), child_index))

            block_maps.append(block_map)

        # run each list in child_inputs through the relevant block's bulk_to_python
        # to obtain child_outputs
        child_outputs = {
            block_type: self.child_blocks[block_type].bulk_to_python(child_input_list)
            for block_type, child_input_list in child_inputs.items()
        }

        # for each stream, go through the block map, picking out the appropriately-indexed
        # value from the relevant list in child_outputs
        return [
            StreamValue(
                self,
                [
                    (block_type, child_outputs[block_type][child_index], id)
                    for block_type, id, child_index in block_map
                ],
                is_lazy=False,
            )
            for block_map in block_maps
        ]

    def get_prep_value(self, value):
        if not value:
            # Falsy values (including None, empty string, empty list, and
            # empty StreamValue) become an empty stream
            return []
        else:
            # value is a StreamValue - delegate to its get_prep_value() method
            # (which has special-case handling for lazy StreamValues to avoid useless
            # round-trips to the full data representation and back)
            return value.get_prep_value()

    def get_form_state(self, value):
        if not value:
            return []
        else:
            return [
                {
                    "type": child.block.name,
                    "value": child.block.get_form_state(child.value),
                    "id": child.id,
                }
                for child in value
            ]

    def get_api_representation(self, value, context=None):
        if value is None:
            # treat None as identical to an empty stream
            return []

        return [
            {
                "type": child.block.name,
                "value": child.block.get_api_representation(
                    child.value, context=context
                ),
                "id": child.id,
            }
            for child in value  # child is a StreamChild instance
        ]

    def render_basic(self, value, context=None):
        return format_html_join(
            "\n",
            '<div class="block-{1}">{0}</div>',
            [(child.render(context=context), child.block_type) for child in value],
        )

    def get_searchable_content(self, value):
        content = []

        for child in value:
            content.extend(child.block.get_searchable_content(child.value))

        return content

    def deconstruct(self):
        """
        Always deconstruct StreamBlock instances as if they were plain StreamBlocks with all of the
        field definitions passed to the constructor - even if in reality this is a subclass of StreamBlock
        with the fields defined declaratively, or some combination of the two.

        This ensures that the field definitions get frozen into migrations, rather than leaving a reference
        to a custom subclass in the user's models.py that may or may not stick around.
        """
        path = "wagtail.blocks.StreamBlock"
        args = [list(self.child_blocks.items())]
        kwargs = self._constructor_kwargs
        return (path, args, kwargs)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        for name, child_block in self.child_blocks.items():
            errors.extend(child_block.check(**kwargs))
            errors.extend(child_block._check_name(**kwargs))

        return errors

    class Meta:
        # No icon specified here, because that depends on the purpose that the
        # block is being used for. Feel encouraged to specify an icon in your
        # descendant block type
        icon = "placeholder"
        default = []
        required = True
        form_classname = None
        min_num = None
        max_num = None
        block_counts = {}
        collapsed = False

    MUTABLE_META_ATTRIBUTES = [
        "required",
        "min_num",
        "max_num",
        "block_counts",
        "collapsed",
    ]


class StreamBlock(BaseStreamBlock, metaclass=DeclarativeSubBlocksMetaclass):
    pass


class StreamValue(MutableSequence):
    """
    Custom type used to represent the value of a StreamBlock; behaves as a sequence of BoundBlocks
    (which keep track of block types in a way that the values alone wouldn't).
    """

    class StreamChild(BoundBlock):
        """
        Iterating over (or indexing into) a StreamValue returns instances of StreamChild.
        These are wrappers for the individual data items in the stream, extending BoundBlock
        (which keeps track of the data item's corresponding Block definition object, and provides
        the `render` method to render itself with a template) with an `id` property (a UUID
        assigned to the item - this is managed by the enclosing StreamBlock and is not a property
        of blocks in general) and a `block_type` property.
        """

        def __init__(self, *args, **kwargs):
            self.id = kwargs.pop("id")
            super(StreamValue.StreamChild, self).__init__(*args, **kwargs)

        @property
        def block_type(self):
            """
            Syntactic sugar so that we can say child.block_type instead of child.block.name.
            (This doesn't belong on BoundBlock itself because the idea of block.name denoting
            the child's "type" ('heading', 'paragraph' etc) is unique to StreamBlock, and in the
            wider context people are liable to confuse it with the block class (CharBlock etc).
            """
            return self.block.name

        def get_prep_value(self):
            return {
                "type": self.block_type,
                "value": self.block.get_prep_value(self.value),
                "id": self.id,
            }

        def _as_tuple(self):
            if self.id:
                return (self.block.name, self.value, self.id)
            else:
                return (self.block.name, self.value)

    class RawDataView(MutableSequence):
        """
        Internal helper class to present the stream data in raw JSONish format. For backwards
        compatibility with old code that manipulated StreamValue.stream_data, this is considered
        mutable to some extent, with the proviso that once the BoundBlock representation has been
        accessed, any changes to fields within raw data will not propagate back to the BoundBlock
        and will not be saved back when calling get_prep_value.
        """

        def __init__(self, stream_value):
            self.stream_value = stream_value

        def __getitem__(self, i):
            item = self.stream_value._raw_data[i]
            if item is None:
                # reconstruct raw data from the bound block
                item = self.stream_value._bound_blocks[i].get_prep_value()
                self.stream_value._raw_data[i] = item

            return item

        def __len__(self):
            return len(self.stream_value._raw_data)

        def __setitem__(self, i, item):
            self.stream_value._raw_data[i] = item
            # clear the cached bound_block for this item
            self.stream_value._bound_blocks[i] = None

        def __delitem__(self, i):
            # same as deletion on the stream itself - delete both the raw and bound_block data
            del self.stream_value[i]

        def insert(self, i, item):
            self.stream_value._raw_data.insert(i, item)
            self.stream_value._bound_blocks.insert(i, None)

        def __repr__(self):
            return repr(list(self))

    def __init__(self, stream_block, stream_data, is_lazy=False, raw_text=None):
        """
        Construct a StreamValue linked to the given StreamBlock,
        with child values given in stream_data.

        Passing is_lazy=True means that stream_data is raw JSONish data as stored
        in the database, and needs to be converted to native values
        (using block.to_python()) when accessed. In this mode, stream_data is a
        list of dicts, each containing 'type' and 'value' keys.

        Passing is_lazy=False means that stream_data consists of immediately usable
        native values. In this mode, stream_data is a list of (type_name, value)
        or (type_name, value, id) tuples.

        raw_text exists solely as a way of representing StreamField content that is
        not valid JSON; this may legitimately occur if an existing text field is
        migrated to a StreamField. In this situation we return a blank StreamValue
        with the raw text accessible under the `raw_text` attribute, so that migration
        code can be rewritten to convert it as desired.
        """
        self.stream_block = (
            stream_block  # the StreamBlock object that handles this value
        )
        self.is_lazy = is_lazy
        self.raw_text = raw_text

        if is_lazy:
            # store raw stream data in _raw_data; on retrieval it will be converted to a native
            # value (via block.to_python) and wrapped as a StreamValue, and cached in _bound_blocks.
            self._raw_data = stream_data
            self._bound_blocks = [None] * len(stream_data)
        else:
            # store native stream data in _bound_blocks; on serialization it will be converted to
            # a JSON-ish representation via block.get_prep_value.
            self._raw_data = [None] * len(stream_data)
            self._bound_blocks = [
                self._construct_stream_child(item) for item in stream_data
            ]

    def _construct_stream_child(self, item):
        """
        Create a StreamChild instance from a (type, value, id) or (type, value) tuple,
        or return item if it's already a StreamChild
        """
        if isinstance(item, StreamValue.StreamChild):
            return item

        try:
            type_name, value, block_id = item
        except ValueError:
            type_name, value = item
            block_id = None

        block_def = self.stream_block.child_blocks[type_name]
        return StreamValue.StreamChild(block_def, value, id=block_id)

    def __getitem__(self, i):
        if isinstance(i, slice):
            start, stop, step = i.indices(len(self._bound_blocks))
            return [self[j] for j in range(start, stop, step)]

        if self._bound_blocks[i] is None:
            raw_value = self._raw_data[i]
            self._prefetch_blocks(raw_value["type"])

        return self._bound_blocks[i]

    def __setitem__(self, i, item):
        self._bound_blocks[i] = self._construct_stream_child(item)

    def __delitem__(self, i):
        del self._bound_blocks[i]
        del self._raw_data[i]

    def insert(self, i, item):
        self._bound_blocks.insert(i, self._construct_stream_child(item))
        self._raw_data.insert(i, None)

    @cached_property
    def raw_data(self):
        return StreamValue.RawDataView(self)

    def _prefetch_blocks(self, type_name):
        """
        Populate _bound_blocks with all items in this stream of type `type_name` that exist in
        _raw_data but do not already exist in _bound_blocks.

        Fetching is done via the block's bulk_to_python method, so that database lookups are
        batched into a single query where possible.
        """
        child_block = self.stream_block.child_blocks[type_name]
        # create a mapping of all the child blocks matching the given block type,
        # mapping (index within the stream) => (raw block value)
        raw_values = OrderedDict(
            (i, raw_item["value"])
            for i, raw_item in enumerate(self._raw_data)
            if raw_item["type"] == type_name and self._bound_blocks[i] is None
        )
        # pass the raw block values to bulk_to_python as a list
        converted_values = child_block.bulk_to_python(raw_values.values())

        # reunite the converted values with their stream indexes, along with the block ID
        # if one exists
        for i, value in zip(raw_values.keys(), converted_values):
            self._bound_blocks[i] = StreamValue.StreamChild(
                child_block, value, id=self._raw_data[i].get("id")
            )

    def get_prep_value(self):
        prep_value = []

        for i, item in enumerate(self._bound_blocks):
            if item:
                # Convert the native value back into raw JSONish data
                if not item.id:
                    item.id = str(uuid.uuid4())

                prep_value.append(item.get_prep_value())
            else:
                # item has not been converted to a BoundBlock, so its _raw_data entry is
                # still usable (but ensure it has an ID before returning it)

                raw_item = self._raw_data[i]
                if not raw_item.get("id"):
                    raw_item["id"] = str(uuid.uuid4())

                prep_value.append(raw_item)

        return prep_value

    def __eq__(self, other):
        if not isinstance(other, StreamValue) or len(other) != len(self):
            return False

        # scan both lists for non-matching items
        for i in range(0, len(self)):
            if self._bound_blocks[i] is None and other._bound_blocks[i] is None:
                # compare raw values as a shortcut to save the conversion step
                if self._raw_data[i] != other._raw_data[i]:
                    return False
            else:
                this_item = self[i]
                other_item = other[i]
                if (
                    this_item.block_type != other_item.block_type
                    or this_item.id != other_item.id
                    or this_item.value != other_item.value
                ):
                    return False

        return True

    def __len__(self):
        return len(self._bound_blocks)

    def __repr__(self):
        return "<%s %r>" % (type(self).__name__, list(self))

    def render_as_block(self, context=None):
        return self.stream_block.render(self, context=context)

    def __html__(self):
        return self.stream_block.render(self)

    def __str__(self):
        return self.__html__()


class StreamBlockAdapter(Adapter):
    js_constructor = "wagtail.blocks.StreamBlock"

    def js_args(self, block):
        meta = {
            "label": block.label,
            "required": block.required,
            "icon": block.meta.icon,
            "classname": block.meta.form_classname,
            "maxNum": block.meta.max_num,
            "minNum": block.meta.min_num,
            "blockCounts": block.meta.block_counts,
            "collapsed": block.meta.collapsed,
            "strings": {
                "MOVE_UP": _("Move up"),
                "MOVE_DOWN": _("Move down"),
                "DUPLICATE": _("Duplicate"),
                "DELETE": _("Delete"),
                "ADD": _("Add"),
            },
        }
        help_text = getattr(block.meta, "help_text", None)
        if help_text:
            meta["helpText"] = help_text
            meta["helpIcon"] = get_help_icon()

        return [
            block.name,
            block.grouped_child_blocks(),
            {
                name: child_block.get_form_state(child_block.get_default())
                for name, child_block in block.child_blocks.items()
            },
            meta,
        ]

    @cached_property
    def media(self):
        return forms.Media(
            js=[
                versioned_static("wagtailadmin/js/telepath/blocks.js"),
            ]
        )


register(StreamBlockAdapter(), StreamBlock)



import os

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404, redirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.decorators import method_decorator
from django.utils.http import urlencode
from django.utils.translation import gettext as _
from django.views.generic import TemplateView

from wagtail.admin import messages
from wagtail.admin.auth import PermissionPolicyChecker
from wagtail.admin.forms.search import SearchForm
from wagtail.admin.models import popular_tags_for_model
from wagtail.admin.views.pages.utils import get_valid_next_url_from_request
from wagtail.images import get_image_model
from wagtail.images.exceptions import InvalidFilterSpecError
from wagtail.images.forms import URLGeneratorForm, get_image_form
from wagtail.images.models import Filter, SourceImageIOError
from wagtail.images.permissions import permission_policy
from wagtail.images.utils import generate_signature
from wagtail.models import Collection, Site
from wagtail.search import index as search_index

permission_checker = PermissionPolicyChecker(permission_policy)

INDEX_PAGE_SIZE = getattr(settings, "WAGTAILIMAGES_INDEX_PAGE_SIZE", 20)
USAGE_PAGE_SIZE = getattr(settings, "WAGTAILIMAGES_USAGE_PAGE_SIZE", 20)


class BaseListingView(TemplateView):
    @method_decorator(permission_checker.require_any("add", "change", "delete"))
    def get(self, request):
        return super().get(request)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Get images (filtered by user permission)
        images = (
            permission_policy.instances_user_has_any_permission_for(
                self.request.user, ["change", "delete"]
            )
            .order_by("-created_at")
            .select_related("collection")
        )

        # Search
        query_string = None
        if "q" in self.request.GET:
            self.form = SearchForm(self.request.GET, placeholder=_("Search images"))
            if self.form.is_valid():
                query_string = self.form.cleaned_data["q"]

                images = images.search(query_string)
        else:
            self.form = SearchForm(placeholder=_("Search images"))

        # Filter by collection
        self.current_collection = None
        collection_id = self.request.GET.get("collection_id")
        if collection_id:
            try:
                self.current_collection = Collection.objects.get(id=collection_id)
                images = images.filter(collection=self.current_collection)
            except (ValueError, Collection.DoesNotExist):
                pass

        # Filter by tag
        self.current_tag = self.request.GET.get("tag")
        if self.current_tag:
            try:
                images = images.filter(tags__name=self.current_tag)
            except (AttributeError):
                self.current_tag = None

        paginator = Paginator(images, per_page=INDEX_PAGE_SIZE)
        images = paginator.get_page(self.request.GET.get("p"))

        next_url = reverse("wagtailimages:index")
        request_query_string = self.request.META.get("QUERY_STRING")
        if request_query_string:
            next_url += "?" + request_query_string

        context.update(
            {
                "images": images,
                "query_string": query_string,
                "is_searching": bool(query_string),
                "next": next_url,
            }
        )

        return context


class IndexView(BaseListingView):
    template_name = "wagtailimages/images/index.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        collections = permission_policy.collections_user_has_any_permission_for(
            self.request.user, ["add", "change"]
        )
        if len(collections) < 2:
            collections = None

        Image = get_image_model()

        context.update(
            {
                "search_form": self.form,
                "popular_tags": popular_tags_for_model(get_image_model()),
                "current_tag": self.current_tag,
                "collections": collections,
                "current_collection": self.current_collection,
                "user_can_add": permission_policy.user_has_permission(
                    self.request.user, "add"
                ),
                "app_label": Image._meta.app_label,
                "model_name": Image._meta.model_name,
            }
        )
        return context


class ListingResultsView(BaseListingView):
    template_name = "wagtailimages/images/results.html"


@permission_checker.require("change")
def edit(request, image_id):
    Image = get_image_model()
    ImageForm = get_image_form(Image)

    image = get_object_or_404(Image, id=image_id)

    if not permission_policy.user_has_permission_for_instance(
        request.user, "change", image
    ):
        raise PermissionDenied

    next_url = get_valid_next_url_from_request(request)

    if request.method == "POST":
        original_file = image.file
        form = ImageForm(request.POST, request.FILES, instance=image, user=request.user)
        if form.is_valid():
            if "file" in form.changed_data:
                # Set new image file size
                image.file_size = image.file.size

                # Set new image file hash
                image.file.seek(0)
                image._set_file_hash(image.file.read())
                image.file.seek(0)

            form.save()

            if "file" in form.changed_data:
                # if providing a new image file, delete the old one and all renditions.
                # NB Doing this via original_file.delete() clears the file field,
                # which definitely isn't what we want...
                original_file.storage.delete(original_file.name)
                image.renditions.all().delete()

            # Reindex the image to make sure all tags are indexed
            search_index.insert_or_update_object(image)

            edit_url = reverse("wagtailimages:edit", args=(image.id,))
            redirect_url = "wagtailimages:index"
            if next_url:
                edit_url = f"{edit_url}?{urlencode({'next': next_url})}"
                redirect_url = next_url

            messages.success(
                request,
                _("Image '{0}' updated.").format(image.title),
                buttons=[messages.button(edit_url, _("Edit again"))],
            )
            return redirect(redirect_url)
        else:
            messages.error(request, _("The image could not be saved due to errors."))
    else:
        form = ImageForm(instance=image, user=request.user)

    # Check if we should enable the frontend url generator
    try:
        reverse("wagtailimages_serve", args=("foo", "1", "bar"))
        url_generator_enabled = True
    except NoReverseMatch:
        url_generator_enabled = False

    if image.is_stored_locally():
        # Give error if image file doesn't exist
        if not os.path.isfile(image.file.path):
            messages.error(
                request,
                _(
                    "The source image file could not be found. Please change the source or delete the image."
                ).format(image.title),
                buttons=[
                    messages.button(
                        reverse("wagtailimages:delete", args=(image.id,)), _("Delete")
                    )
                ],
            )

    try:
        filesize = image.get_file_size()
    except SourceImageIOError:
        filesize = None

    return TemplateResponse(
        request,
        "wagtailimages/images/edit.html",
        {
            "image": image,
            "form": form,
            "url_generator_enabled": url_generator_enabled,
            "filesize": filesize,
            "user_can_delete": permission_policy.user_has_permission_for_instance(
                request.user, "delete", image
            ),
            "next": next_url,
        },
    )


def url_generator(request, image_id):
    image = get_object_or_404(get_image_model(), id=image_id)

    if not permission_policy.user_has_permission_for_instance(
        request.user, "change", image
    ):
        raise PermissionDenied

    form = URLGeneratorForm(
        initial={
            "filter_method": "original",
            "width": image.width,
            "height": image.height,
        }
    )

    return TemplateResponse(
        request,
        "wagtailimages/images/url_generator.html",
        {
            "image": image,
            "form": form,
        },
    )


def generate_url(request, image_id, filter_spec):
    # Get the image
    Image = get_image_model()
    try:
        image = Image.objects.get(id=image_id)
    except Image.DoesNotExist:
        return JsonResponse({"error": "Cannot find image."}, status=404)

    # Check if this user has edit permission on this image
    if not permission_policy.user_has_permission_for_instance(
        request.user, "change", image
    ):
        return JsonResponse(
            {"error": "You do not have permission to generate a URL for this image."},
            status=403,
        )

    # Parse the filter spec to make sure its valid
    try:
        Filter(spec=filter_spec).operations
    except InvalidFilterSpecError:
        return JsonResponse({"error": "Invalid filter spec."}, status=400)

    # Generate url
    signature = generate_signature(image_id, filter_spec)
    url = reverse("wagtailimages_serve", args=(signature, image_id, filter_spec))

    # Get site root url
    try:
        site_root_url = Site.objects.get(is_default_site=True).root_url
    except Site.DoesNotExist:
        site_root_url = Site.objects.first().root_url

    # Generate preview url
    preview_url = reverse("wagtailimages:preview", args=(image_id, filter_spec))

    return JsonResponse(
        {"url": site_root_url + url, "preview_url": preview_url}, status=200
    )


def preview(request, image_id, filter_spec):
    image = get_object_or_404(get_image_model(), id=image_id)

    try:
        response = HttpResponse()
        image = Filter(spec=filter_spec).run(image, response)
        response["Content-Type"] = "image/" + image.format_name
        return response
    except InvalidFilterSpecError:
        return HttpResponse(
            "Invalid filter spec: " + filter_spec, content_type="text/plain", status=400
        )


@permission_checker.require("delete")
def delete(request, image_id):
    image = get_object_or_404(get_image_model(), id=image_id)

    if not permission_policy.user_has_permission_for_instance(
        request.user, "delete", image
    ):
        raise PermissionDenied

    next_url = get_valid_next_url_from_request(request)

    if request.method == "POST":
        image.delete()
        messages.success(request, _("Image '{0}' deleted.").format(image.title))
        return redirect(next_url) if next_url else redirect("wagtailimages:index")

    return TemplateResponse(
        request,
        "wagtailimages/images/confirm_delete.html",
        {
            "image": image,
            "next": next_url,
        },
    )


@permission_checker.require("add")
def add(request):
    ImageModel = get_image_model()
    ImageForm = get_image_form(ImageModel)

    if request.method == "POST":
        image = ImageModel(uploaded_by_user=request.user)
        form = ImageForm(request.POST, request.FILES, instance=image, user=request.user)
        if form.is_valid():
            # Set image file size
            image.file_size = image.file.size

            # Set image file hash
            image.file.seek(0)
            image._set_file_hash(image.file.read())
            image.file.seek(0)

            form.save()

            # Reindex the image to make sure all tags are indexed
            search_index.insert_or_update_object(image)

            messages.success(
                request,
                _("Image '{0}' added.").format(image.title),
                buttons=[
                    messages.button(
                        reverse("wagtailimages:edit", args=(image.id,)), _("Edit")
                    )
                ],
            )
            return redirect("wagtailimages:index")
        else:
            messages.error(request, _("The image could not be created due to errors."))
    else:
        form = ImageForm(user=request.user)

    return TemplateResponse(
        request,
        "wagtailimages/images/add.html",
        {
            "form": form,
        },
    )


def usage(request, image_id):
    image = get_object_or_404(get_image_model(), id=image_id)

    paginator = Paginator(image.get_usage(), per_page=USAGE_PAGE_SIZE)
    used_by = paginator.get_page(request.GET.get("p"))

    return TemplateResponse(
        request, "wagtailimages/images/usage.html", {"image": image, "used_by": used_by}
    )



import datetime
from io import BytesIO
from unittest import mock

from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core import checks
from django.test import TestCase
from django.test.utils import override_settings
from django.utils.timezone import make_aware
from openpyxl import load_workbook

from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.admin.panels import FieldPanel, TabbedInterface
from wagtail.contrib.modeladmin.helpers.search import DjangoORMSearchHandler
from wagtail.images.models import Image
from wagtail.images.tests.utils import get_test_image_file
from wagtail.models import Locale, ModelLogEntry, Page
from wagtail.test.modeladmintest.models import (
    Author,
    Book,
    Publisher,
    RelatedLink,
    Token,
    TranslatableBook,
)
from wagtail.test.modeladmintest.wagtail_hooks import BookModelAdmin
from wagtail.test.utils import WagtailTestUtils


class TestBookIndexView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

        img = Image.objects.create(
            title="LOTR cover",
            file=get_test_image_file(),
        )
        book = Book.objects.get(title="The Lord of the Rings")
        book.cover_image = img
        book.save()

    def get(self, **params):
        return self.client.get("/admin/modeladmintest/book/", params)

    def test_thumbnail_image_col_header_text(self):
        response = self.get()

        # check thumb_col_header_text is correctly used
        self.assertContains(
            response,
            '<th scope="col" class="column-admin_thumb">The cover</th>',
            html=True,
        )

    def test_simple(self):
        response = self.get()

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context["result_count"], 4)

        # The result count content is shown in the header
        self.assertContains(
            response, '<span class="result-count">4 out of 4</span>', html=True
        )

        # User has add permission
        self.assertIs(response.context["user_can_create"], True)

    def test_csv_export(self):
        # Export the whole queryset
        response = self.get(export="csv")

        self.assertEqual(response.status_code, 200)
        # Check attachment is present and named correctly using the modeladmin export_filename
        self.assertEqual(
            response.get("content-disposition"),
            'attachment; filename="books-export.csv"',
        )

        # Check response - all books should be in it
        data_lines = response.getvalue().decode().split("\n")
        self.assertEqual(data_lines[0], "Title,Author,Author Date Of Birth\r")
        self.assertEqual(
            data_lines[1], "Charlie and the Chocolate Factory,Roald Dahl,1916-09-13\r"
        )
        self.assertEqual(
            data_lines[2], "The Chronicles of Narnia,Roald Dahl,1898-11-29\r"
        )
        self.assertEqual(data_lines[3], "The Hobbit,J. R. R. Tolkien,1892-01-03\r")
        self.assertEqual(
            data_lines[4], "The Lord of the Rings,J. R. R. Tolkien,1892-01-03\r"
        )

    def test_xlsx_export(self):
        # Export the whole queryset
        response = self.get(export="xlsx")

        self.assertEqual(response.status_code, 200)
        # Check attachment is present and named correctly using the modeladmin export_filename
        self.assertEqual(
            response.get("content-disposition"),
            'attachment; filename="books-export.xlsx"',
        )

        # Check response - all books should be in it
        workbook_data = response.getvalue()
        worksheet = load_workbook(filename=BytesIO(workbook_data))["Sheet1"]
        cell_array = [[cell.value for cell in row] for row in worksheet.rows]
        self.assertEqual(cell_array[0], ["Title", "Author", "Author Date Of Birth"])
        self.assertEqual(
            cell_array[1],
            ["Charlie and the Chocolate Factory", "Roald Dahl", "1916-09-13"],
        )
        self.assertEqual(
            cell_array[2], ["The Chronicles of Narnia", "Roald Dahl", "1898-11-29"]
        )
        self.assertEqual(
            cell_array[3], ["The Hobbit", "J. R. R. Tolkien", "1892-01-03"]
        )
        self.assertEqual(
            cell_array[4], ["The Lord of the Rings", "J. R. R. Tolkien", "1892-01-03"]
        )
        self.assertEqual(len(cell_array), 5)

    def test_tr_attributes(self):
        response = self.get()

        # Charlie & The Chocolate factory should be in the list with the
        # `data-author_yob` and `data-object_pk` attributes added
        self.assertContains(response, 'data-author-yob="1916"')
        self.assertContains(response, 'data-object-pk="3"')

        # There should be two odd rows and two even ones, and 'book' should be
        # added to the `class` attribute for every one.
        self.assertContains(response, 'class="book odd"', count=2)
        self.assertContains(response, 'class="book even"', count=2)

    def test_filter(self):
        # Filter by author 1 (JRR Tolkien)
        response = self.get(author__id__exact=1)

        self.assertEqual(response.status_code, 200)

        # JRR Tolkien has two books in the test data
        self.assertEqual(response.context["result_count"], 2)

        # The result count content is shown in the header
        self.assertContains(
            response, '<span class="result-count">2 out of 4</span>', html=True
        )

        for book in response.context["object_list"]:
            self.assertEqual(book.author_id, 1)

    def test_filtered_csv_export(self):
        # Filter by author 1 (JRR Tolkien) and export the current selection
        response = self.get(author__id__exact=1, export="csv")

        # Check response - only books by JRR Tolkien should be in it
        self.assertEqual(response.status_code, 200)
        data_lines = response.getvalue().decode().split("\n")
        self.assertEqual(data_lines[0], "Title,Author,Author Date Of Birth\r")
        self.assertEqual(data_lines[1], "The Hobbit,J. R. R. Tolkien,1892-01-03\r")
        self.assertEqual(
            data_lines[2], "The Lord of the Rings,J. R. R. Tolkien,1892-01-03\r"
        )
        self.assertEqual(data_lines[3], "")

    def test_search_indexed(self):
        response = self.get(q="of")

        self.assertEqual(response.status_code, 200)

        # There are two books where the title contains 'of'
        self.assertEqual(response.context["result_count"], 2)

        # The result count content is shown in the header
        self.assertContains(
            response, '<span class="result-count">2 out of 4</span>', html=True
        )

    def test_search_form_present(self):
        # Test the backend search handler allows the search form to render
        response = self.get()

        self.assertContains(response, '<input id="id_q"')

    def test_search_form_absent(self):
        # DjangoORMSearchHandler + no search_fields, search form should be absent
        with mock.patch.object(
            BookModelAdmin, "search_handler_class", DjangoORMSearchHandler
        ):
            response = self.get()

            self.assertNotContains(response, '<input id="id_q"')

    def test_ordering(self):
        response = self.get(o="0.1")

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context["result_count"], 4)

    def test_paging(self):
        # should be corrected to just the first page, as there aren't enough
        # objects to make up more than one page
        response = self.get(p=9)

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context["result_count"], 4)

        # Should raise a ValueError that gets caught during initialisation
        response = self.get(p="notaninteger")

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context["result_count"], 4)


class TestAuthorIndexView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self, **params):
        return self.client.get("/admin/modeladmintest/author/", params)

    def test_search(self):
        response = self.get(q="Roald Dahl")

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context["result_count"], 2)

        # The result count content is shown in the header
        self.assertContains(
            response, '<span class="result-count">2 out of 5</span>', html=True
        )

    def test_col_extra_class_names(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        test_html = """
            <td class="field-first_book for-author-1 title">The Lord of the Rings</td>
        """
        self.assertContains(response, test_html, html=True)

    def test_col_extra_attributes(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        test_html = """
            <td class="field-last_book title" data-for_author="1">The Hobbit</td>
        """
        self.assertContains(response, test_html, html=True)


@override_settings(WAGTAIL_I18N_ENABLED=True)
class TestTranslatableBookIndexView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self, **params):
        return self.client.get("/admin/modeladmintest/translatablebook/", params)

    def test_simple(self):
        response = self.get()

        self.assertEqual(response.status_code, 200)

        # There are two books in the test data
        self.assertEqual(response.context["result_count"], 2)

        # Check the locale filter is there
        expected = """
        <ul>
            <li class="selected">
            <a href="?">All</a></li>
            <li>
            <a href="?locale__id__exact=1">English</a></li>
            <li>
            <a href="?locale__id__exact=2">French</a></li>
        </ul>"""
        self.assertContains(response, expected, html=True)

    @override_settings(WAGTAIL_I18N_ENABLED=False)
    def test_locale_selector_not_present_when_i18n_disabled(self):
        response = self.get()

        self.assertNotContains(
            response, '<a href="?locale__id__exact=2">French</a>', html=True
        )

    def test_filter(self):
        # Filter by locale 2 (fr)
        response = self.get(locale__id__exact=2)

        self.assertEqual(response.status_code, 200)

        # Locale fr has one book in the test data
        self.assertEqual(response.context["result_count"], 1)

        for book in response.context["object_list"]:
            self.assertEqual(book.locale_id, 2)

        self.assertContains(response, "Le Seigneur des anneaux", html=True)


class TestCreateView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self):
        return self.client.get("/admin/modeladmintest/book/create/")

    def post(self, post_data):
        return self.client.post("/admin/modeladmintest/book/create/", post_data)

    def test_simple(self):
        response = self.get()

        self.assertEqual(response.status_code, 200)

    def test_create(self):
        response = self.post(
            {
                "title": "George's Marvellous Medicine",
                "author": 2,
            }
        )
        # Should redirect back to index
        self.assertRedirects(response, "/admin/modeladmintest/book/")

        # Check that the book was created
        self.assertEqual(
            Book.objects.filter(title="George's Marvellous Medicine").count(), 1
        )

        response = self.client.get("/admin/modeladmintest/publisher/create/")
        self.assertIn("name", response.content.decode("UTF-8"))
        self.assertNotIn("headquartered_in", response.content.decode("UTF-8"))
        self.assertEqual(list(response.context["form"].fields), ["name"])
        self.client.post(
            "/admin/modeladmintest/publisher/create/", {"name": "Sharper Collins"}
        )
        publisher = Publisher.objects.get(name="Sharper Collins")
        self.assertIsNone(publisher.headquartered_in)

    def test_post_invalid(self):
        initial_book_count = Book.objects.count()

        response = self.post(
            {
                "title": "",
                "author": 2,
            }
        )
        final_book_count = Book.objects.count()

        self.assertEqual(response.status_code, 200)
        # Check that the book was not created
        self.assertEqual(initial_book_count, final_book_count)

        # Check that a form error was raised
        self.assertFormError(response, "form", "title", "This field is required.")
        self.assertContains(
            response,
            """<p class="error-message"><span>This field is required.</span></p>""",
            count=1,
            html=True,
        )

    def test_exclude_passed_to_extract_panel_definitions(self):
        path_to_form_fields_exclude_property = (
            "wagtail.contrib.modeladmin.options.ModelAdmin.form_fields_exclude"
        )
        with mock.patch(
            "wagtail.contrib.modeladmin.options.extract_panel_definitions_from_model_class"
        ) as m:
            with mock.patch(
                path_to_form_fields_exclude_property, new_callable=mock.PropertyMock
            ) as mock_form_fields_exclude:
                mock_form_fields_exclude.return_value = ["123"]

                self.get()
                self.assertTrue(mock_form_fields_exclude.called)
                m.assert_called_with(
                    Book, exclude=mock_form_fields_exclude.return_value
                )

    def test_clean_form_once(self):
        with mock.patch(
            "wagtail.test.modeladmintest.wagtail_hooks.PublisherModelAdminForm.clean"
        ) as mock_form_clean:
            response = self.client.post(
                "/admin/modeladmintest/publisher/create/", {"name": ""}
            )
            self.assertEqual(response.status_code, 200)

            mock_form_clean.assert_called_once()

    def test_create_view_with_multifieldpanel(self):
        # https://github.com/wagtail/wagtail/issues/6413
        response = self.client.get("/admin/modeladmintest/relatedlink/create/")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            "/admin/modeladmintest/relatedlink/create/",
            {
                "title": "Homepage",
                "link": Page.objects.filter(depth=2).first().id,
            },
        )
        # Should redirect back to index
        self.assertRedirects(response, "/admin/modeladmintest/relatedlink/")

        # Check that the link was created
        self.assertEqual(RelatedLink.objects.filter(title="Homepage").count(), 1)

    def test_prepopulated_field_data_in_context(self):
        response = self.get()
        self.assertIn(
            'data-prepopulated-fields="[{&quot;id&quot;: &quot;#id_title&quot;, &quot;name&quot;: &quot;title&quot;, &quot;dependency_ids&quot;: [&quot;#id_author&quot;], &quot;dependency_list&quot;: [&quot;author&quot;], &quot;maxLength&quot;: 255, &quot;allowUnicode&quot;: false}]"',
            response.content.decode("UTF-8"),
        )


@override_settings(WAGTAIL_I18N_ENABLED=True)
class TestTranslatableCreateView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self, **params):
        return self.client.get("/admin/modeladmintest/translatablebook/create/", params)

    def test_simple(self):
        response = self.get(locale="fr")

        self.assertEqual(response.status_code, 200)

        # Check that the locale select exists and is set correctly
        expected = '<a href="javascript:void(0)" aria-label="French" class="c-dropdown__button u-btn-current w-no-underline">'
        self.assertContains(response, expected)

        # Check that the other locale link is right
        expected = """
        <a href="/admin/modeladmintest/translatablebook/create/?locale=en" aria-label="English" class="u-link is-live w-no-underline">
            English
        </a>"""
        self.assertContains(response, expected, html=True)


class TestInspectView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

        img = Image.objects.create(
            title="LOTR cover",
            file=get_test_image_file(),
        )
        book = Book.objects.get(title="The Lord of the Rings")
        book.cover_image = img
        book.save()

    def get_for_author(self, author_id):
        return self.client.get("/admin/modeladmintest/author/inspect/%d/" % author_id)

    def get_for_book(self, book_id):
        return self.client.get("/admin/modeladmintest/book/inspect/%d/" % book_id)

    def test_author_simple(self):
        response = self.get_for_author(1)
        self.assertEqual(response.status_code, 200)

    def test_author_name_present(self):
        """
        The author name should appear twice. Once in the header, and once
        more in the field listing
        """
        response = self.get_for_author(1)
        self.assertContains(response, "J. R. R. Tolkien", 2)

    def test_author_dob_not_present(self):
        """
        The date of birth shouldn't appear, because the field wasn't included
        in the `inspect_view_fields` list
        """
        response = self.get_for_author(1)
        self.assertNotContains(response, "1892")

    def test_book_simple(self):
        response = self.get_for_book(1)
        self.assertEqual(response.status_code, 200)

    def test_book_title_present(self):
        """
        The book title should appear once only, in the header, as 'title'
        was added to the `inspect_view_fields_ignore` list
        """
        response = self.get_for_book(1)
        self.assertContains(response, "The Lord of the Rings", 1)

    def test_book_author_present(self):
        """
        The author name should appear, because 'author' is not in
        `inspect_view_fields_ignore` and should be returned by the
        `get_inspect_view_fields` method.
        """
        response = self.get_for_book(1)
        self.assertContains(response, "J. R. R. Tolkien", 1)

    def test_non_existent(self):
        response = self.get_for_book(100)
        self.assertEqual(response.status_code, 404)


class TestEditView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.user = self.login()
        ModelLogEntry.objects.create(
            content_type=ContentType.objects.get_for_model(Book),
            label="The Lord of the Rings",
            action="wagtail.create",
            timestamp=make_aware(datetime.datetime(2021, 9, 30, 10, 1, 0)),
            object_id="1",
        )

    def get(self, book_id):
        return self.client.get("/admin/modeladmintest/book/edit/%d/" % book_id)

    def post(self, book_id, post_data):
        return self.client.post(
            "/admin/modeladmintest/book/edit/%d/" % book_id, post_data
        )

    def test_simple(self):
        response = self.get(1)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "The Lord of the Rings")

        # "Last updated" timestamp should be present
        self.assertContains(
            response, 'data-wagtail-tooltip="Sept. 30, 2021, 10:01 a.m."'
        )
        # History link should be present
        self.assertContains(response, 'href="/admin/modeladmintest/book/history/1/"')

        url_finder = AdminURLFinder(self.user)
        expected_url = "/admin/modeladmintest/book/edit/1/"
        self.assertEqual(url_finder.get_edit_url(Book.objects.get(id=1)), expected_url)

    def test_non_existent(self):
        response = self.get(100)

        self.assertEqual(response.status_code, 404)

    def test_edit(self):
        response = self.post(
            1,
            {
                "title": "The Lady of the Rings",
                "author": 1,
            },
        )

        # Should redirect back to index
        self.assertRedirects(response, "/admin/modeladmintest/book/")

        # Check that the book was updated
        self.assertEqual(Book.objects.get(id=1).title, "The Lady of the Rings")

    def test_post_invalid(self):
        response = self.post(
            1,
            {
                "title": "",
                "author": 1,
            },
        )

        self.assertEqual(response.status_code, 200)

        # Check that the title was not updated
        self.assertEqual(Book.objects.get(id=1).title, "The Lord of the Rings")

        # Check that a form error was raised
        self.assertFormError(response, "form", "title", "This field is required.")
        self.assertContains(
            response,
            """<p class="error-message"><span>This field is required.</span></p>""",
            count=1,
            html=True,
        )

    def test_exclude_passed_to_extract_panel_definitions(self):
        path_to_form_fields_exclude_property = (
            "wagtail.contrib.modeladmin.options.ModelAdmin.form_fields_exclude"
        )
        with mock.patch(
            "wagtail.contrib.modeladmin.options.extract_panel_definitions_from_model_class"
        ) as m:
            with mock.patch(
                path_to_form_fields_exclude_property, new_callable=mock.PropertyMock
            ) as mock_form_fields_exclude:
                mock_form_fields_exclude.return_value = ["123"]

                self.get(1)
                self.assertTrue(mock_form_fields_exclude.called)
                m.assert_called_with(
                    Book, exclude=mock_form_fields_exclude.return_value
                )

    def test_clean_form_once(self):
        with mock.patch(
            "wagtail.test.modeladmintest.wagtail_hooks.PublisherModelAdminForm.clean"
        ) as mock_form_clean:
            publisher = Publisher.objects.create(name="Sharper Collins")

            response = self.client.post(
                "/admin/modeladmintest/publisher/edit/%d/" % publisher.pk, {"name": ""}
            )
            self.assertEqual(response.status_code, 200)

            mock_form_clean.assert_called_once()

    def test_edit_view_with_multifieldpanel(self):
        # https://github.com/wagtail/wagtail/issues/6413
        link = RelatedLink.objects.create(
            title="Homepage", link=Page.objects.filter(depth=2).first()
        )
        response = self.client.get(
            "/admin/modeladmintest/relatedlink/edit/%d/" % link.id
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            "/admin/modeladmintest/relatedlink/edit/%d/" % link.id,
            {
                "title": "Homepage edited",
                "link": Page.objects.filter(depth=2).first().id,
            },
        )
        # Should redirect back to index
        self.assertRedirects(response, "/admin/modeladmintest/relatedlink/")

        # Check that the link was updated
        self.assertEqual(RelatedLink.objects.filter(title="Homepage edited").count(), 1)


@override_settings(WAGTAIL_I18N_ENABLED=True)
class TestTranslatableBookEditView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self, book_id, **params):
        return self.client.get(
            "/admin/modeladmintest/translatablebook/edit/%d/" % book_id, params
        )

    def test_simple(self):
        book = TranslatableBook.objects.first()
        response = self.get(book.id)
        self.assertEqual(response.status_code, 200)

        # Check the locale switcher isn't there
        self.assertNotContains(response, "English", html=True)

        tbook = book.copy_for_translation(locale=Locale.objects.get(language_code="fr"))
        tbook.save()

        response = self.get(tbook.id)
        self.assertEqual(response.status_code, 200)

        # Check the locale switcher is there
        expected = """
        <a href="/admin/modeladmintest/translatablebook/edit/1/?locale=en" aria-label="English" class="u-link is-live w-no-underline">
            English
        </a>"""
        self.assertContains(response, expected, html=True)


class TestPageSpecificViews(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]
    expected_status_code = 404

    def setUp(self):
        self.login()

    def test_choose_parent(self):
        response = self.client.get("/admin/modeladmintest/book/choose_parent/")
        self.assertEqual(response.status_code, self.expected_status_code)


class TestConfirmDeleteView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self, book_id):
        return self.client.get("/admin/modeladmintest/book/delete/%d/" % book_id)

    def post(self, book_id):
        return self.client.post("/admin/modeladmintest/book/delete/%d/" % book_id)

    def test_simple(self):
        response = self.get(1)

        self.assertEqual(response.status_code, 200)

    def test_non_existent(self):
        response = self.get(100)

        self.assertEqual(response.status_code, 404)

    def test_post(self):
        response = self.post(1)

        # User redirected to index
        self.assertRedirects(response, "/admin/modeladmintest/book/")

        # Book deleted
        self.assertFalse(Book.objects.filter(id=1).exists())


class TestDeleteViewWithProtectedRelation(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def get(self, author_id):
        return self.client.get("/admin/modeladmintest/author/delete/%d/" % author_id)

    def post(self, author_id):
        return self.client.post("/admin/modeladmintest/author/delete/%d/" % author_id)

    def test_get_with_dependent_object(self):
        response = self.get(1)

        self.assertEqual(response.status_code, 200)

    def test_get_without_dependent_object(self):
        response = self.get(4)

        self.assertEqual(response.status_code, 200)

    def test_post_with_dependent_object(self):
        response = self.post(1)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, "'J. R. R. Tolkien' is currently referenced by other objects"
        )
        self.assertContains(response, "<li><b>Book:</b> The Lord of the Rings</li>")

        # Author not deleted
        self.assertTrue(Author.objects.filter(id=1).exists())

    def test_post_without_dependent_object(self):
        response = self.post(4)

        # User redirected to index
        self.assertRedirects(response, "/admin/modeladmintest/author/")

        # Author deleted
        self.assertFalse(Author.objects.filter(id=4).exists())

    def test_post_with_1to1_dependent_object(self):
        response = self.post(5)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, "'Harper Lee' is currently referenced by other objects"
        )
        self.assertContains(
            response, "<li><b>Solo Book:</b> To Kill a Mockingbird</li>"
        )

        # Author not deleted
        self.assertTrue(Author.objects.filter(id=5).exists())


class TestDeleteViewModelReprPrimary(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def test_delete(self):
        response = self.client.post("/admin/modeladmintest/token/delete/boom/")
        self.assertEqual(response.status_code, 302)


class TestEditorAccess(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        # Create a user
        user = self.create_user(username="test2", password="password")
        user.groups.add(Group.objects.get(pk=2))
        self.user = user
        # Login
        self.login(username="test2", password="password")

        return user

    def test_index_permitted(self):
        response = self.client.get("/admin/modeladmintest/book/")
        self.assertRedirects(response, "/admin/")

    def test_inpspect_permitted(self):
        response = self.client.get("/admin/modeladmintest/book/inspect/2/")
        self.assertRedirects(response, "/admin/")

    def test_create_permitted(self):
        response = self.client.get("/admin/modeladmintest/book/create/")
        self.assertRedirects(response, "/admin/")

    def test_edit_permitted(self):
        response = self.client.get("/admin/modeladmintest/book/edit/2/")
        self.assertRedirects(response, "/admin/")

        url_finder = AdminURLFinder(self.user)
        self.assertIsNone(url_finder.get_edit_url(Book.objects.get(id=2)))

    def test_delete_get_permitted(self):
        response = self.client.get("/admin/modeladmintest/book/delete/2/")
        self.assertRedirects(response, "/admin/")

    def test_delete_post_permitted(self):
        response = self.client.post("/admin/modeladmintest/book/delete/2/")
        self.assertRedirects(response, "/admin/")


class TestHistoryView(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()
        ModelLogEntry.objects.create(
            content_type=ContentType.objects.get_for_model(Book),
            label="The Lord of the Rings",
            action="wagtail.create",
            timestamp=make_aware(datetime.datetime(2021, 9, 30, 10, 1, 0)),
            object_id="1",
        )

    def test_simple(self):
        response = self.client.get("/admin/modeladmintest/book/history/1/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "<td>Created</td>", html=True)
        self.assertContains(
            response,
            '<div class="human-readable-date" title="Sept. 30, 2021, 10:01 a.m.">',
        )


class TestQuoting(TestCase, WagtailTestUtils):
    fixtures = ["modeladmintest_test.json"]
    expected_status_code = 200

    def setUp(self):
        self.login()
        self.tok_reg = Token.objects.create(key="RegularName")
        self.tok_irr = Token.objects.create(key="Irregular_Name")

    def test_action_links(self):
        response = self.client.get("/admin/modeladmintest/token/")

        self.assertContains(
            response, 'href="/admin/modeladmintest/token/edit/RegularName/"'
        )
        self.assertContains(
            response, 'href="/admin/modeladmintest/token/delete/RegularName/"'
        )
        self.assertContains(
            response, 'href="/admin/modeladmintest/token/edit/Irregular_5FName/"'
        )
        self.assertContains(
            response, 'href="/admin/modeladmintest/token/delete/Irregular_5FName/"'
        )

        response = self.client.get("/admin/modeladmintest/token/edit/Irregular_5FName/")
        self.assertEqual(response.status_code, 200)
        response = self.client.get(
            "/admin/modeladmintest/token/delete/Irregular_5FName/"
        )
        self.assertEqual(response.status_code, 200)


class TestHeaderBreadcrumbs(TestCase, WagtailTestUtils):
    """
    Test that the breadcrumbs region is inserted before the
    </header> tag for potential future regression.
    See https://github.com/wagtail/wagtail/issues/3889
    """

    fixtures = ["modeladmintest_test.json"]

    def setUp(self):
        self.login()

    def test_choose_inspect_model(self):
        response = self.client.get("/admin/modeladmintest/author/inspect/1/")

        # check correct templates were used
        self.assertTemplateUsed(response, "modeladmin/includes/breadcrumb.html")
        self.assertTemplateUsed(response, "wagtailadmin/shared/header.html")

        # check that home breadcrumb link exists
        expected = """
            <li class="breadcrumb-item home">
                <a href="/admin/" class="breadcrumb-link">
                    <svg class="icon icon-home home_icon" aria-hidden="true">
                        <use href="#icon-home"></use>
                    </svg>
                    <span class="visuallyhidden">Home</span>
                    <svg class="icon icon-arrow-right arrow_right_icon" aria-hidden="true">
                        <use href="#icon-arrow-right"></use>
                    </svg>
                </a>
            </li>
        """
        self.assertContains(response, expected, html=True)

        # check that the breadcrumbs are before the first header closing tag
        content_str = str(response.content)
        position_of_header_close = content_str.index("</header>")
        position_of_breadcrumbs = content_str.index('<ul class="breadcrumb">')
        self.assertGreater(position_of_header_close, position_of_breadcrumbs)


class TestPanelConfigurationChecks(TestCase, WagtailTestUtils):
    def setUp(self):
        self.warning_id = "wagtailadmin.W002"

        def get_checks_result():
            # run checks only with the 'panels' tag
            checks_result = checks.run_checks(tags=["panels"])
            return [
                warning for warning in checks_result if warning.id == self.warning_id
            ]

        self.get_checks_result = get_checks_result

    def test_model_with_single_tabbed_panel_only(self):
        Publisher.content_panels = [FieldPanel("name"), FieldPanel("headquartered_in")]

        warning = checks.Warning(
            "Publisher.content_panels will have no effect on modeladmin editing",
            hint="""Ensure that Publisher uses `panels` instead of `content_panels`\
or set up an `edit_handler` if you want a tabbed editing interface.
There are no default tabs on non-Page models so there will be no\
 Content tab for the content_panels to render in.""",
            obj=Publisher,
            id="wagtailadmin.W002",
        )

        checks_results = self.get_checks_result()

        self.assertIn(warning, checks_results)

        # clean up for future checks
        delattr(Publisher, "content_panels")

    def test_model_with_two_tabbed_panels_only(self):
        Publisher.settings_panels = [FieldPanel("name")]
        Publisher.promote_panels = [FieldPanel("headquartered_in")]

        warning_1 = checks.Warning(
            "Publisher.promote_panels will have no effect on modeladmin editing",
            hint="""Ensure that Publisher uses `panels` instead of `promote_panels`\
or set up an `edit_handler` if you want a tabbed editing interface.
There are no default tabs on non-Page models so there will be no\
 Promote tab for the promote_panels to render in.""",
            obj=Publisher,
            id="wagtailadmin.W002",
        )

        warning_2 = checks.Warning(
            "Publisher.settings_panels will have no effect on modeladmin editing",
            hint="""Ensure that Publisher uses `panels` instead of `settings_panels`\
or set up an `edit_handler` if you want a tabbed editing interface.
There are no default tabs on non-Page models so there will be no\
 Settings tab for the settings_panels to render in.""",
            obj=Publisher,
            id="wagtailadmin.W002",
        )

        checks_results = self.get_checks_result()

        self.assertIn(warning_1, checks_results)
        self.assertIn(warning_2, checks_results)

        # clean up for future checks
        delattr(Publisher, "settings_panels")
        delattr(Publisher, "promote_panels")

    def test_model_with_single_tabbed_panel_and_edit_handler(self):
        Publisher.content_panels = [FieldPanel("name"), FieldPanel("headquartered_in")]
        Publisher.edit_handler = TabbedInterface(Publisher.content_panels)

        # no errors should occur
        self.assertEqual(self.get_checks_result(), [])

        # clean up for future checks
        delattr(Publisher, "content_panels")
        delattr(Publisher, "edit_handler")



from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic.base import View

from wagtail import hooks
from wagtail.admin.auth import PermissionPolicyChecker
from wagtail.admin.forms.search import SearchForm
from wagtail.admin.modal_workflow import render_modal_workflow
from wagtail.admin.ui.tables import Column, DateColumn, Table, TitleColumn
from wagtail.documents import get_document_model
from wagtail.documents.forms import get_document_form
from wagtail.documents.permissions import permission_policy
from wagtail.search import index as search_index

permission_checker = PermissionPolicyChecker(permission_policy)


def get_document_chosen_response(request, document):
    """
    helper function: given a document, return the modal workflow response that returns that
    document back to the calling page
    """
    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={
            "step": "document_chosen",
            "result": {
                "id": document.id,
                "title": document.title,
                "url": document.url,
                "filename": document.filename,
                "edit_link": reverse("wagtaildocs:edit", args=(document.id,)),
            },
        },
    )


class DownloadColumn(Column):
    cell_template_name = "wagtaildocs/tables/download_cell.html"

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["download_url"] = instance.url
        return context


class BaseChooseView(View):
    def get(self, request):
        Document = get_document_model()

        if permission_policy.user_has_permission(request.user, "add"):
            DocumentForm = get_document_form(Document)
            self.uploadform = DocumentForm(
                user=request.user, prefix="document-chooser-upload"
            )
        else:
            self.uploadform = None

        documents = permission_policy.instances_user_has_any_permission_for(
            request.user, ["choose"]
        )

        # allow hooks to modify the queryset
        for hook in hooks.get_hooks("construct_document_chooser_queryset"):
            documents = hook(documents, request)

        self.q = None
        self.is_searching = False

        self.collection_id = request.GET.get("collection_id")
        if self.collection_id:
            documents = documents.filter(collection=self.collection_id)
        self.documents_exist = documents.exists()

        if "q" in request.GET:
            self.searchform = SearchForm(request.GET)
            if self.searchform.is_valid():
                self.q = self.searchform.cleaned_data["q"]

                documents = documents.search(self.q)
                self.is_searching = True
        else:
            self.searchform = SearchForm()

        if not self.is_searching:
            documents = documents.order_by("-created_at")

        paginator = Paginator(documents, per_page=10)
        self.documents = paginator.get_page(request.GET.get("p"))

        self.collections = permission_policy.collections_user_has_permission_for(
            request.user, "choose"
        )
        if len(self.collections) < 2:
            self.collections = None

        columns = [
            TitleColumn(
                "title",
                label=_("Title"),
                url_name="wagtaildocs:document_chosen",
                link_classname="document-choice",
            ),
            DownloadColumn("filename", label=_("File")),
            DateColumn("created_at", label=_("Created"), width="16%"),
        ]

        if self.collections:
            columns.insert(2, Column("collection", label=_("Collection")))

        self.table = Table(columns, self.documents)

        return self.render_to_response()

    def get_context_data(self):
        return {
            "documents": self.documents,
            "documents_exist": self.documents_exist,
            "table": self.table,
            "uploadform": self.uploadform,
            "query_string": self.q,
            "searchform": self.searchform,
            "collections": self.collections,
            "is_searching": self.is_searching,
            "collection_id": self.collection_id,
        }

    def render_to_response(self):
        raise NotImplementedError()


class ChooseView(BaseChooseView):
    def render_to_response(self):
        return render_modal_workflow(
            self.request,
            "wagtaildocs/chooser/chooser.html",
            None,
            self.get_context_data(),
            json_data={
                "step": "chooser",
                "error_label": _("Server Error"),
                "error_message": _(
                    "Report this error to your website administrator with the following information:"
                ),
                "tag_autocomplete_url": reverse("wagtailadmin_tag_autocomplete"),
            },
        )


class ChooseResultsView(BaseChooseView):
    def render_to_response(self):
        return TemplateResponse(
            self.request, "wagtaildocs/chooser/results.html", self.get_context_data()
        )


def document_chosen(request, document_id):
    document = get_object_or_404(get_document_model(), id=document_id)
    return get_document_chosen_response(request, document)


@permission_checker.require("add")
def chooser_upload(request):
    Document = get_document_model()
    DocumentForm = get_document_form(Document)

    if request.method == "POST":
        document = Document(uploaded_by_user=request.user)
        form = DocumentForm(
            request.POST,
            request.FILES,
            instance=document,
            user=request.user,
            prefix="document-chooser-upload",
        )

        if form.is_valid():
            document.file_size = document.file.size

            # Set new document file hash
            document.file.seek(0)
            document._set_file_hash(document.file.read())
            document.file.seek(0)

            form.save()

            # Reindex the document to make sure all tags are indexed
            search_index.insert_or_update_object(document)
            return get_document_chosen_response(request, document)
    else:
        form = DocumentForm(user=request.user, prefix="document-chooser-upload")

    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={
            "step": "reshow_upload_form",
            "htmlFragment": render_to_string(
                "wagtaildocs/chooser/upload_form.html", {"form": form}, request
            ),
        },
    )



import hashlib

from django.conf import settings
from django.utils.http import urlencode

from wagtail.compat import AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME

delete_user_perm = "{0}.delete_{1}".format(
    AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME.lower()
)


def user_can_delete_user(current_user, user_to_delete):
    if not current_user.has_perm(delete_user_perm):
        return False

    if current_user == user_to_delete:
        # users may not delete themselves
        return False

    if user_to_delete.is_superuser and not current_user.is_superuser:
        # ordinary users may not delete superusers
        return False

    return True


def get_gravatar_url(email, size=50):
    default = "mm"
    size = (
        int(size) * 2
    )  # requested at retina size by default and scaled down at point of use with css
    gravatar_provider_url = getattr(
        settings, "WAGTAIL_GRAVATAR_PROVIDER_URL", "//www.gravatar.com/avatar"
    )

    if (not email) or (gravatar_provider_url is None):
        return None

    gravatar_url = "{gravatar_provider_url}/{hash}?{params}".format(
        gravatar_provider_url=gravatar_provider_url.rstrip("/"),
        hash=hashlib.md5(email.lower().encode("utf-8")).hexdigest(),
        params=urlencode({"s": size, "d": default}),
    )

    return gravatar_url



from django.test import TestCase, override_settings
from django.urls import reverse

from wagtail.models import Page
from wagtail.test.utils import WagtailTestUtils


class TestLoginView(TestCase, WagtailTestUtils):
    fixtures = ["test.json"]

    def setUp(self):
        self.user = self.create_test_user()
        self.homepage = Page.objects.get(url_path="/home/")

    def test_success_redirect(self):
        response = self.client.post(
            reverse("wagtailadmin_login"),
            {
                "username": "test@email.com",
                "password": "password",
            },
        )
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_success_redirect_honour_redirect_get_parameter(self):
        homepage_admin_url = reverse("wagtailadmin_pages:edit", args=[self.homepage.pk])
        login_url = reverse("wagtailadmin_login") + "?next={}".format(
            homepage_admin_url
        )
        response = self.client.post(
            login_url,
            {
                "username": "test@email.com",
                "password": "password",
            },
        )
        self.assertRedirects(response, homepage_admin_url)

    def test_success_redirect_honour_redirect_post_parameter(self):
        homepage_admin_url = reverse("wagtailadmin_pages:edit", args=[self.homepage.pk])
        response = self.client.post(
            reverse("wagtailadmin_login"),
            {
                "username": "test@email.com",
                "password": "password",
                "next": homepage_admin_url,
            },
        )
        self.assertRedirects(response, homepage_admin_url)

    def test_already_authenticated_redirect(self):
        self.login(username="test@email.com", password="password")

        response = self.client.get(reverse("wagtailadmin_login"))
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_already_authenticated_redirect_honour_redirect_get_parameter(self):
        self.login(username="test@email.com", password="password")

        homepage_admin_url = reverse("wagtailadmin_pages:edit", args=[self.homepage.pk])
        login_url = reverse("wagtailadmin_login") + "?next={}".format(
            homepage_admin_url
        )
        response = self.client.get(login_url)
        self.assertRedirects(response, homepage_admin_url)

    @override_settings(LANGUAGE_CODE="de")
    def test_language_code(self):
        response = self.client.get(reverse("wagtailadmin_login"))
        self.assertContains(response, '<html lang="de" dir="ltr">')

    @override_settings(LANGUAGE_CODE="he")
    def test_bidi_language_changes_dir_attribute(self):
        response = self.client.get(reverse("wagtailadmin_login"))
        self.assertContains(response, '<html lang="he" dir="rtl">')

    @override_settings(
        WAGTAILADMIN_USER_LOGIN_FORM="wagtail.admin.tests.test_forms.CustomLoginForm"
    )
    def test_login_page_renders_extra_fields(self):
        response = self.client.get(reverse("wagtailadmin_login"))
        self.assertContains(
            response, '<input type="text" name="captcha" required id="id_captcha">'
        )

    def test_session_expire_on_browser_close(self):
        self.client.post(
            reverse("wagtailadmin_login"),
            {
                "username": "test@email.com",
                "password": "password",
            },
        )
        self.assertTrue(self.client.session.get_expire_at_browser_close())

    @override_settings(SESSION_COOKIE_AGE=7)
    def test_session_expiry_remember(self):
        self.client.post(
            reverse("wagtailadmin_login"),
            {"username": "test@email.com", "password": "password", "remember": True},
        )
        self.assertFalse(self.client.session.get_expire_at_browser_close())
        self.assertEqual(self.client.session.get_expiry_age(), 7)



from wagtail.coreutils import (  # noqa
    SCRIPT_RE,
    SLUGIFY_RE,
    WAGTAIL_APPEND_SLASH,
    InvokeViaAttributeShortcut,
    accepts_kwarg,
    camelcase_to_underscore,
    cautious_slugify,
    escape_script,
    find_available_slug,
    get_content_languages,
    get_content_type_label,
    get_model_string,
    get_supported_content_language_variant,
    multigetattr,
    resolve_model_string,
    safe_snake_case,
    string_to_ascii,
)



from django.test import TestCase

from wagtail.templatetags.wagtailcore_tags import richtext


class TestTemplateTag(TestCase):
    def test_no_contrib_legacy_richtext_no_wrapper(self):
        self.assertEqual(richtext("Foo"), "Foo")

    def test_contrib_legacy_richtext_renders_wrapper(self):
        with self.modify_settings(
            INSTALLED_APPS={"prepend": "wagtail.contrib.legacy.richtext"}
        ):
            self.assertEqual(richtext("Foo"), """<div class="rich-text">Foo</div>""")



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0002_initial_data"),
    ]

    operations = [
        migrations.CreateModel(
            name="Redirect",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        auto_created=True,
                        serialize=False,
                        primary_key=True,
                    ),
                ),
                (
                    "old_path",
                    models.CharField(
                        verbose_name="Redirect from",
                        max_length=255,
                        unique=True,
                        db_index=True,
                    ),
                ),
                (
                    "is_permanent",
                    models.BooleanField(
                        verbose_name="Permanent",
                        default=True,
                        help_text="""Recommended. Permanent redirects \
                    ensure search engines forget the old page (the 'Redirect from') and index the new page instead.""",
                    ),
                ),
                (
                    "redirect_link",
                    models.URLField(blank=True, verbose_name="Redirect to any URL"),
                ),
                (
                    "redirect_page",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        blank=True,
                        null=True,
                        verbose_name="Redirect to a page",
                        to="wagtailcore.Page",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        blank=True,
                        to="wagtailcore.Site",
                        editable=False,
                        null=True,
                        related_name="redirects",
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
    ]



from io import BytesIO
from unittest.mock import patch

from django.test import TestCase, override_settings

from wagtail import hooks
from wagtail.images import image_operations
from wagtail.images.exceptions import InvalidFilterSpecError
from wagtail.images.image_operations import TransformOperation
from wagtail.images.models import Filter, Image
from wagtail.images.tests.utils import (
    get_test_image_file,
    get_test_image_file_jpeg,
    get_test_image_file_webp,
)


class DummyImageTransform:
    """
    This class pretends to be a ImageTransform but instead, it records
    the operations that have been performed on it
    """

    def __init__(self, size):
        self._check_size(size)
        self.size = size
        self.operations = []

    def clone(self):
        clone = DummyImageTransform(self.size)
        clone.operations = self.operations.copy()
        return clone

    def resize(self, size):
        """
        Change the image size, stretching the transform to make it fit the new size.
        """
        self._check_size(size)
        clone = self.clone()
        clone.operations.append(("resize", size))
        clone.size = size
        return clone

    def crop(self, rect):
        """
        Crop the image to the specified rect.
        """
        self._check_size(tuple(rect.size))
        clone = self.clone()
        clone.operations.append(("crop", tuple(rect)))
        clone.size = tuple(rect.size)
        return clone

    @staticmethod
    def _check_size(size):
        if (
            not isinstance(size, tuple)
            or len(size) != 2
            or int(size[0]) != size[0]
            or int(size[1]) != size[1]
        ):
            raise TypeError("Image size must be a 2-tuple of integers")

        if size[0] < 1 or size[1] < 1:
            raise ValueError("Image width and height must both be 1 or greater")


class ImageTransformOperationTestCase(TestCase):
    operation_class = None
    filter_spec_tests = []
    filter_spec_error_tests = []
    run_tests = []
    norun_tests = []

    @classmethod
    def make_filter_spec_test(cls, filter_spec, expected_output):
        def test_filter_spec(self):
            operation = self.operation_class(*filter_spec.split("-"))

            # Check the attributes are set correctly
            for attr, value in expected_output.items():
                self.assertEqual(getattr(operation, attr), value)

        test_filter_spec.__name__ = str("test_filter_%s" % filter_spec)
        return test_filter_spec

    @classmethod
    def make_filter_spec_error_test(cls, filter_spec):
        def test_filter_spec_error(self):
            self.assertRaises(
                InvalidFilterSpecError, self.operation_class, *filter_spec.split("-")
            )

        test_filter_spec_error.__name__ = str(
            "test_filter_%s_raises_%s" % (filter_spec, InvalidFilterSpecError.__name__)
        )
        return test_filter_spec_error

    @classmethod
    def make_run_test(cls, filter_spec, image_kwargs, expected_output):
        def test_run(self):
            image = Image(**image_kwargs)

            # Make operation
            operation = self.operation_class(*filter_spec.split("-"))

            # Make context
            context = DummyImageTransform((image.width, image.height))

            # Run
            context = operation.run(context, image)

            # Check
            self.assertEqual(context.operations, expected_output)

        test_run.__name__ = str("test_run_%s" % filter_spec)
        return test_run

    @classmethod
    def make_norun_test(cls, filter_spec, image_kwargs):
        def test_norun(self):
            image = Image(**image_kwargs)

            # Make operation
            operation = self.operation_class(*filter_spec.split("-"))

            # Make operation recorder
            context = DummyImageTransform((image.width, image.height))

            # Attempt (and hopefully fail) to run
            with self.assertRaises(ValueError):
                operation.run(context, image)

        test_norun.__name__ = str("test_norun_%s" % filter_spec)
        return test_norun

    @classmethod
    def setup_test_methods(cls):
        if cls.operation_class is None:
            return

        # Filter spec tests
        for args in cls.filter_spec_tests:
            filter_spec_test = cls.make_filter_spec_test(*args)
            setattr(cls, filter_spec_test.__name__, filter_spec_test)

        # Filter spec error tests
        for filter_spec in cls.filter_spec_error_tests:
            filter_spec_error_test = cls.make_filter_spec_error_test(filter_spec)
            setattr(cls, filter_spec_error_test.__name__, filter_spec_error_test)

        # Running tests
        for args in cls.run_tests:
            run_test = cls.make_run_test(*args)
            setattr(cls, run_test.__name__, run_test)

        # Runtime error tests
        for args in cls.norun_tests:
            norun_test = cls.make_norun_test(*args)
            setattr(cls, norun_test.__name__, norun_test)


class TestFillOperation(ImageTransformOperationTestCase):
    operation_class = image_operations.FillOperation

    filter_spec_tests = [
        ("fill-800x600", {"width": 800, "height": 600, "crop_closeness": 0}),
        ("hello-800x600", {"width": 800, "height": 600, "crop_closeness": 0}),
        ("fill-800x600-c0", {"width": 800, "height": 600, "crop_closeness": 0}),
        ("fill-800x600-c100", {"width": 800, "height": 600, "crop_closeness": 1}),
        ("fill-800x600-c50", {"width": 800, "height": 600, "crop_closeness": 0.5}),
        ("fill-800x600-c1000", {"width": 800, "height": 600, "crop_closeness": 1}),
        ("fill-800000x100", {"width": 800000, "height": 100, "crop_closeness": 0}),
    ]

    filter_spec_error_tests = [
        "fill",
        "fill-800",
        "fill-abc",
        "fill-800xabc",
        "fill-800x600-",
        "fill-800x600x10",
        "fill-800x600-d100",
    ]

    run_tests = [
        # Basic usage
        (
            "fill-800x600",
            {"width": 1000, "height": 1000},
            [
                ("crop", (0, 125, 1000, 875)),
                ("resize", (800, 600)),
            ],
        ),
        # Basic usage with an oddly-sized original image
        # This checks for a rounding precision issue (#968)
        (
            "fill-200x200",
            {"width": 539, "height": 720},
            [
                ("crop", (0, 90, 539, 630)),
                ("resize", (200, 200)),
            ],
        ),
        # Closeness shouldn't have any effect when used without a focal point
        (
            "fill-800x600-c100",
            {"width": 1000, "height": 1000},
            [
                ("crop", (0, 125, 1000, 875)),
                ("resize", (800, 600)),
            ],
        ),
        # Should always crop towards focal point. Even if no closeness is set
        (
            "fill-80x60",
            {
                "width": 1000,
                "height": 1000,
                "focal_point_x": 1000,
                "focal_point_y": 500,
                "focal_point_width": 0,
                "focal_point_height": 0,
            },
            [
                # Crop the largest possible crop box towards the focal point
                ("crop", (0, 125, 1000, 875)),
                # Resize it down to final size
                ("resize", (80, 60)),
            ],
        ),
        # Should crop as close as possible without upscaling
        (
            "fill-80x60-c100",
            {
                "width": 1000,
                "height": 1000,
                "focal_point_x": 1000,
                "focal_point_y": 500,
                "focal_point_width": 0,
                "focal_point_height": 0,
            },
            [
                # Crop as close as possible to the focal point
                ("crop", (920, 470, 1000, 530)),
                # No need to resize, crop should've created an 80x60 image
            ],
        ),
        # Ditto with a wide image
        # Using a different filter so method name doesn't clash
        (
            "fill-100x60-c100",
            {
                "width": 2000,
                "height": 1000,
                "focal_point_x": 2000,
                "focal_point_y": 500,
                "focal_point_width": 0,
                "focal_point_height": 0,
            },
            [
                # Crop to the right hand side
                ("crop", (1900, 470, 2000, 530)),
            ],
        ),
        # Make sure that the crop box never enters the focal point
        (
            "fill-50x50-c100",
            {
                "width": 2000,
                "height": 1000,
                "focal_point_x": 1000,
                "focal_point_y": 500,
                "focal_point_width": 100,
                "focal_point_height": 20,
            },
            [
                # Crop a 100x100 box around the entire focal point
                ("crop", (950, 450, 1050, 550)),
                # Resize it down to 50x50
                ("resize", (50, 50)),
            ],
        ),
        # Test that the image is never upscaled
        (
            "fill-1000x800",
            {"width": 100, "height": 100},
            [
                ("crop", (0, 10, 100, 90)),
            ],
        ),
        # Test that the crop closeness gets capped to prevent upscaling
        (
            "fill-1000x800-c100",
            {
                "width": 1500,
                "height": 1000,
                "focal_point_x": 750,
                "focal_point_y": 500,
                "focal_point_width": 0,
                "focal_point_height": 0,
            },
            [
                # Crop a 1000x800 square out of the image as close to the
                # focal point as possible. Will not zoom too far in to
                # prevent upscaling
                ("crop", (250, 100, 1250, 900)),
            ],
        ),
        # Test for an issue where a ZeroDivisionError would occur when the
        # focal point size, image size and filter size match
        # See: #797
        (
            "fill-1500x1500-c100",
            {
                "width": 1500,
                "height": 1500,
                "focal_point_x": 750,
                "focal_point_y": 750,
                "focal_point_width": 1500,
                "focal_point_height": 1500,
            },
            [
                # This operation could probably be optimised out
                ("crop", (0, 0, 1500, 1500)),
            ],
        ),
        # A few tests for single pixel images
        (
            "fill-100x100",
            {
                "width": 1,
                "height": 1,
            },
            [
                ("crop", (0, 0, 1, 1)),
            ],
        ),
        # This one once gave a ZeroDivisionError
        (
            "fill-100x150",
            {
                "width": 1,
                "height": 1,
            },
            [
                ("crop", (0, 0, 1, 1)),
            ],
        ),
        (
            "fill-150x100",
            {
                "width": 1,
                "height": 1,
            },
            [
                ("crop", (0, 0, 1, 1)),
            ],
        ),
    ]


TestFillOperation.setup_test_methods()


class TestMinMaxOperation(ImageTransformOperationTestCase):
    operation_class = image_operations.MinMaxOperation

    filter_spec_tests = [
        ("min-800x600", {"method": "min", "width": 800, "height": 600}),
        ("max-800x600", {"method": "max", "width": 800, "height": 600}),
    ]

    filter_spec_error_tests = [
        "min",
        "min-800",
        "min-abc",
        "min-800xabc",
        "min-800x600-",
        "min-800x600-c100",
        "min-800x600x10",
    ]

    run_tests = [
        # Basic usage of min
        (
            "min-800x600",
            {"width": 1000, "height": 1000},
            [
                ("resize", (800, 800)),
            ],
        ),
        # Basic usage of max
        (
            "max-800x600",
            {"width": 1000, "height": 1000},
            [
                ("resize", (600, 600)),
            ],
        ),
        # Resize doesn't try to set zero height
        (
            "max-400x400",
            {"width": 1000, "height": 1},
            [
                ("resize", (400, 1)),
            ],
        ),
        # Resize doesn't try to set zero width
        (
            "max-400x400",
            {"width": 1, "height": 1000},
            [
                ("resize", (1, 400)),
            ],
        ),
    ]


TestMinMaxOperation.setup_test_methods()


class TestWidthHeightOperation(ImageTransformOperationTestCase):
    operation_class = image_operations.WidthHeightOperation

    filter_spec_tests = [
        ("width-800", {"method": "width", "size": 800}),
        ("height-600", {"method": "height", "size": 600}),
    ]

    filter_spec_error_tests = [
        "width",
        "width-800x600",
        "width-abc",
        "width-800-c100",
    ]

    run_tests = [
        # Basic usage of width
        (
            "width-400",
            {"width": 1000, "height": 500},
            [
                ("resize", (400, 200)),
            ],
        ),
        # Basic usage of height
        (
            "height-400",
            {"width": 1000, "height": 500},
            [
                ("resize", (800, 400)),
            ],
        ),
        # Resize doesn't try to set zero height
        (
            "width-400",
            {"width": 1000, "height": 1},
            [
                ("resize", (400, 1)),
            ],
        ),
        # Resize doesn't try to set zero width
        (
            "height-400",
            {"width": 1, "height": 800},
            [
                ("resize", (1, 400)),
            ],
        ),
    ]


TestWidthHeightOperation.setup_test_methods()


class TestScaleOperation(ImageTransformOperationTestCase):
    operation_class = image_operations.ScaleOperation

    filter_spec_tests = [
        ("scale-100", {"method": "scale", "percent": 100}),
        ("scale-50", {"method": "scale", "percent": 50}),
    ]

    filter_spec_error_tests = [
        "scale",
        "scale-800x600",
        "scale-abc",
        "scale-800-c100",
    ]

    run_tests = [
        # Basic almost a no-op of scale
        (
            "scale-100",
            {"width": 1000, "height": 500},
            [
                ("resize", (1000, 500)),
            ],
        ),
        # Basic usage of scale
        (
            "scale-50",
            {"width": 1000, "height": 500},
            [
                ("resize", (500, 250)),
            ],
        ),
        # Rounded usage of scale
        (
            "scale-83.0322",
            {"width": 1000, "height": 500},
            [
                ("resize", (int(1000 * 0.830322), int(500 * 0.830322))),
            ],
        ),
        # Resize doesn't try to set zero height
        (
            "scale-50",
            {"width": 1000, "height": 1},
            [
                ("resize", (500, 1)),
            ],
        ),
        # Resize doesn't try to set zero width
        (
            "scale-50",
            {"width": 1, "height": 500},
            [
                ("resize", (1, 250)),
            ],
        ),
    ]


TestScaleOperation.setup_test_methods()


class TestCacheKey(TestCase):
    def test_cache_key(self):
        image = Image(width=1000, height=1000)
        fil = Filter(spec="max-100x100")
        cache_key = fil.get_cache_key(image)

        self.assertEqual(cache_key, "")

    def test_cache_key_fill_filter(self):
        image = Image(width=1000, height=1000)
        fil = Filter(spec="fill-100x100")
        cache_key = fil.get_cache_key(image)

        self.assertEqual(cache_key, "2e16d0ba")

    def test_cache_key_fill_filter_with_focal_point(self):
        image = Image(
            width=1000,
            height=1000,
            focal_point_width=100,
            focal_point_height=100,
            focal_point_x=500,
            focal_point_y=500,
        )
        fil = Filter(spec="fill-100x100")
        cache_key = fil.get_cache_key(image)

        self.assertEqual(cache_key, "0bbe3b2f")


class DummyOperation(TransformOperation):
    def construct(self):
        pass

    def run_mock(self, context, image):
        pass

    def run(self, context, image):
        self.run_mock(context, image)
        return context


def register_image_operations_hook():
    return [("operation1", DummyOperation), ("operation2", DummyOperation)]


class TestFilter(TestCase):
    @patch.object(DummyOperation, "run_mock")
    @hooks.register_temporarily(
        "register_image_operations", register_image_operations_hook
    )
    def test_runs_operations(self, run_mock):
        fil = Filter(spec="operation1|operation2")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        fil.run(image, BytesIO())

        self.assertEqual(run_mock.call_count, 2)


class TestFormatFilter(TestCase):
    def test_jpeg(self):
        fil = Filter(spec="width-400|format-jpeg")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertEqual(out.format_name, "jpeg")

    def test_png(self):
        fil = Filter(spec="width-400|format-png")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertEqual(out.format_name, "png")

    def test_gif(self):
        fil = Filter(spec="width-400|format-gif")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertEqual(out.format_name, "gif")

    def test_webp(self):
        fil = Filter(spec="width-400|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertEqual(out.format_name, "webp")

    def test_webp_lossless(self):
        fil = Filter(spec="width-400|format-webp-lossless")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        # quality=80 is default for Williw and PIL libs
        save.assert_called_with(f, "WEBP", quality=80, lossless=True)

    def test_invalid(self):
        fil = Filter(spec="width-400|format-foo")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())


class TestJPEGQualityFilter(TestCase):
    def test_default_quality(self):
        fil = Filter(spec="width-400")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "JPEG", quality=85, optimize=True, progressive=True)

    def test_jpeg_quality_filter(self):
        fil = Filter(spec="width-400|jpegquality-40")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "JPEG", quality=40, optimize=True, progressive=True)

    def test_jpeg_quality_filter_invalid(self):
        fil = Filter(spec="width-400|jpegquality-abc")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())

    def test_jpeg_quality_filter_no_value(self):
        fil = Filter(spec="width-400|jpegquality")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())

    def test_jpeg_quality_filter_too_big(self):
        fil = Filter(spec="width-400|jpegquality-101")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())

    @override_settings(WAGTAILIMAGES_JPEG_QUALITY=50)
    def test_jpeg_quality_setting(self):
        fil = Filter(spec="width-400")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "JPEG", quality=50, optimize=True, progressive=True)

    @override_settings(WAGTAILIMAGES_JPEG_QUALITY=50)
    def test_jpeg_quality_filter_overrides_setting(self):
        fil = Filter(spec="width-400|jpegquality-40")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "JPEG", quality=40, optimize=True, progressive=True)


class TestWebPQualityFilter(TestCase):
    def test_default_quality(self):
        fil = Filter(spec="width-400|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "WEBP", quality=85, lossless=False)

    def test_webp_quality_filter(self):
        fil = Filter(spec="width-400|webpquality-40|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "WEBP", quality=40, lossless=False)

    def test_webp_quality_filter_invalid(self):
        fil = Filter(spec="width-400|webpquality-abc|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())

    def test_webp_quality_filter_no_value(self):
        fil = Filter(spec="width-400|webpquality")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())

    def test_webp_quality_filter_too_big(self):
        fil = Filter(spec="width-400|webpquality-101|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )
        self.assertRaises(InvalidFilterSpecError, fil.run, image, BytesIO())

    @override_settings(WAGTAILIMAGES_WEBP_QUALITY=50)
    def test_webp_quality_setting(self):
        fil = Filter(spec="width-400|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "WEBP", quality=50, lossless=False)

    @override_settings(WAGTAILIMAGES_WEBP_QUALITY=50)
    def test_jpeg_quality_filter_overrides_setting(self):
        fil = Filter(spec="width-400|webpquality-40|format-webp")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_jpeg(),
        )

        f = BytesIO()
        with patch("PIL.Image.Image.save") as save:
            fil.run(image, f)

        save.assert_called_with(f, "WEBP", quality=40, lossless=False)


class TestBackgroundColorFilter(TestCase):
    def test_original_has_alpha(self):
        # Checks that the test image we're using has alpha
        fil = Filter(spec="width-400")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertTrue(out.has_alpha())

    def test_3_digit_hex(self):
        fil = Filter(spec="width-400|bgcolor-fff")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertFalse(out.has_alpha())

    def test_6_digit_hex(self):
        fil = Filter(spec="width-400|bgcolor-ffffff")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        out = fil.run(image, BytesIO())

        self.assertFalse(out.has_alpha())

    def test_invalid(self):
        fil = Filter(spec="width-400|bgcolor-foo")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        self.assertRaises(ValueError, fil.run, image, BytesIO())

    def test_invalid_length(self):
        fil = Filter(spec="width-400|bgcolor-1234")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file(),
        )
        self.assertRaises(ValueError, fil.run, image, BytesIO())


class TestWebpFormatConversion(TestCase):
    def test_webp_convert_to_png(self):
        """by default, webp images will be converted to png"""

        fil = Filter(spec="width-400")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_webp(),
        )
        out = fil.run(image, BytesIO())

        self.assertEqual(out.format_name, "png")

    @override_settings(WAGTAILIMAGES_FORMAT_CONVERSIONS={"webp": "webp"})
    def test_override_webp_convert_to_png(self):
        """WAGTAILIMAGES_FORMAT_CONVERSIONS can be overridden to disable webp conversion"""

        fil = Filter(spec="width-400")
        image = Image.objects.create(
            title="Test image",
            file=get_test_image_file_webp(),
        )
        out = fil.run(image, BytesIO())

        self.assertEqual(out.format_name, "webp")



from django.forms import BooleanField, ValidationError
from django.utils.timezone import now
from django.utils.translation import gettext as _

from .models import WagtailAdminModelForm


class CommentReplyForm(WagtailAdminModelForm):
    class Meta:
        fields = ("text",)

    def clean(self):
        cleaned_data = super().clean()
        user = self.for_user

        if not self.instance.pk:
            self.instance.user = user
        elif self.instance.user != user:
            # trying to edit someone else's comment reply
            if any(field for field in self.changed_data):
                # includes DELETION_FIELD_NAME, as users cannot delete each other's individual comment replies
                # if deleting a whole thread, this should be done by deleting the parent Comment instead
                self.add_error(
                    None, ValidationError(_("You cannot edit another user's comment."))
                )
        return cleaned_data


class CommentForm(WagtailAdminModelForm):
    """
    This is designed to be subclassed and have the user overridden to enable user-based validation within the edit handler system
    """

    resolved = BooleanField(required=False)

    class Meta:
        formsets = {
            "replies": {
                "form": CommentReplyForm,
                "inherit_kwargs": ["for_user"],
            }
        }

    def clean(self):
        cleaned_data = super().clean()
        user = self.for_user

        if not self.instance.pk:
            self.instance.user = user
        elif self.instance.user != user:
            # trying to edit someone else's comment
            if (
                any(
                    field
                    for field in self.changed_data
                    if field not in ["resolved", "position", "contentpath"]
                )
                or cleaned_data["contentpath"].split(".")[0]
                != self.instance.contentpath.split(".")[0]
            ):
                # users can resolve each other's base comments and change their positions within a field, or move a comment between blocks in a StreamField
                self.add_error(
                    None, ValidationError(_("You cannot edit another user's comment."))
                )
        return cleaned_data

    def save(self, *args, **kwargs):
        if self.cleaned_data.get("resolved", False):
            if not getattr(self.instance, "resolved_at"):
                self.instance.resolved_at = now()
                self.instance.resolved_by = self.for_user
        else:
            self.instance.resolved_by = None
            self.instance.resolved_at = None
        return super().save(*args, **kwargs)



import json

from django import forms
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail.admin.staticfiles import versioned_static
from wagtail.admin.widgets import AdminChooser
from wagtail.images import get_image_model
from wagtail.images.shortcuts import get_rendition_or_not_found
from wagtail.telepath import register
from wagtail.widget_adapters import WidgetAdapter


class AdminImageChooser(AdminChooser):
    choose_one_text = _("Choose an image")
    choose_another_text = _("Change image")
    link_to_chosen_text = _("Edit this image")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.image_model = get_image_model()

    def get_value_data(self, value):
        if value is None:
            return None
        elif isinstance(value, self.image_model):
            image = value
        else:  # assume image ID
            image = self.image_model.objects.get(pk=value)

        preview_image = get_rendition_or_not_found(image, "max-165x165")

        return {
            "id": image.pk,
            "title": image.title,
            "preview": {
                "url": preview_image.url,
                "width": preview_image.width,
                "height": preview_image.height,
            },
            "edit_url": reverse("wagtailimages:edit", args=[image.id]),
        }

    def render_html(self, name, value_data, attrs):
        value_data = value_data or {}
        original_field_html = super().render_html(name, value_data.get("id"), attrs)

        return render_to_string(
            "wagtailimages/widgets/image_chooser.html",
            {
                "widget": self,
                "original_field_html": original_field_html,
                "attrs": attrs,
                "value": bool(
                    value_data
                ),  # only used by chooser.html to identify blank values
                "title": value_data.get("title", ""),
                "preview": value_data.get("preview", {}),
                "edit_url": value_data.get("edit_url", ""),
            },
        )

    def render_js_init(self, id_, name, value_data):
        return "createImageChooser({0});".format(json.dumps(id_))

    @property
    def media(self):
        return forms.Media(
            js=[
                versioned_static("wagtailimages/js/image-chooser-modal.js"),
                versioned_static("wagtailimages/js/image-chooser.js"),
            ]
        )


class ImageChooserAdapter(WidgetAdapter):
    js_constructor = "wagtail.images.widgets.ImageChooser"

    def js_args(self, widget):
        return [
            widget.render_html("__NAME__", None, attrs={"id": "__ID__"}),
            widget.id_for_label("__ID__"),
        ]

    @cached_property
    def media(self):
        return forms.Media(
            js=[
                versioned_static("wagtailimages/js/image-chooser-telepath.js"),
            ]
        )


register(ImageChooserAdapter(), AdminImageChooser)



from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest
from django.template import engines
from django.test import TestCase

from wagtail.models import PAGE_TEMPLATE_VAR, Page, Site
from wagtail.test.utils import WagtailTestUtils


class TestCoreJinja(TestCase, WagtailTestUtils):
    def setUp(self):
        self.engine = engines["jinja2"]

        self.user = self.create_superuser(
            username="test", email="test@email.com", password="password"
        )
        self.homepage = Page.objects.get(id=2)

    def render(self, string, context=None, request_context=True):
        if context is None:
            context = {}

        template = self.engine.from_string(string)
        return template.render(context)

    def dummy_request(self, user=None):
        site = Site.objects.get(is_default_site=True)

        request = HttpRequest()
        request.META["HTTP_HOST"] = site.hostname
        request.META["SERVER_PORT"] = site.port
        request.user = user or AnonymousUser()
        return request

    def test_userbar(self):
        content = self.render(
            "{{ wagtailuserbar() }}",
            {
                PAGE_TEMPLATE_VAR: self.homepage,
                "request": self.dummy_request(self.user),
            },
        )
        self.assertIn("<!-- Wagtail user bar embed code -->", content)

    def test_userbar_anonymous_user(self):
        content = self.render(
            "{{ wagtailuserbar() }}",
            {PAGE_TEMPLATE_VAR: self.homepage, "request": self.dummy_request()},
        )

        # Make sure nothing was rendered
        self.assertEqual(content, "")



from collections import OrderedDict

from django.conf import settings
from rest_framework.pagination import BasePagination
from rest_framework.response import Response

from .utils import BadRequestError


class WagtailPagination(BasePagination):
    def paginate_queryset(self, queryset, request, view=None):
        limit_max = getattr(settings, "WAGTAILAPI_LIMIT_MAX", 20)

        try:
            offset = int(request.GET.get("offset", 0))
            if offset < 0:
                raise ValueError()
        except ValueError:
            raise BadRequestError("offset must be a positive integer")

        try:
            limit_default = 20 if not limit_max else min(20, limit_max)
            limit = int(request.GET.get("limit", limit_default))
            if limit < 0:
                raise ValueError()
        except ValueError:
            raise BadRequestError("limit must be a positive integer")

        if limit_max and limit > limit_max:
            raise BadRequestError("limit cannot be higher than %d" % limit_max)

        start = offset
        stop = offset + limit

        self.view = view
        self.total_count = queryset.count()
        return queryset[start:stop]

    def get_paginated_response(self, data):
        data = OrderedDict(
            [
                (
                    "meta",
                    OrderedDict(
                        [
                            ("total_count", self.total_count),
                        ]
                    ),
                ),
                ("items", data),
            ]
        )
        return Response(data)



from django.conf import settings
from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.http import urlencode
from django.utils.translation import gettext as _
from django.views.generic.base import View

from wagtail import hooks
from wagtail.admin.auth import PermissionPolicyChecker
from wagtail.admin.forms.search import SearchForm
from wagtail.admin.modal_workflow import render_modal_workflow
from wagtail.admin.models import popular_tags_for_model
from wagtail.images import get_image_model
from wagtail.images.formats import get_image_format
from wagtail.images.forms import ImageInsertionForm, get_image_form
from wagtail.images.permissions import permission_policy
from wagtail.images.utils import find_image_duplicates
from wagtail.search import index as search_index

permission_checker = PermissionPolicyChecker(permission_policy)

CHOOSER_PAGE_SIZE = getattr(settings, "WAGTAILIMAGES_CHOOSER_PAGE_SIZE", 12)


def get_image_result_data(image):
    """
    helper function: given an image, return the json data to pass back to the
    image chooser panel
    """
    preview_image = image.get_rendition("max-165x165")

    return {
        "id": image.id,
        "edit_link": reverse("wagtailimages:edit", args=(image.id,)),
        "title": image.title,
        "preview": {
            "url": preview_image.url,
            "width": preview_image.width,
            "height": preview_image.height,
        },
    }


class BaseChooseView(View):
    def get(self, request):
        self.image_model = get_image_model()

        images = permission_policy.instances_user_has_any_permission_for(
            request.user, ["choose"]
        ).order_by("-created_at")

        # allow hooks to modify the queryset
        for hook in hooks.get_hooks("construct_image_chooser_queryset"):
            images = hook(images, request)

        collection_id = request.GET.get("collection_id")
        if collection_id:
            images = images.filter(collection=collection_id)

        self.is_searching = False
        self.q = None

        if "q" in request.GET:
            self.search_form = SearchForm(request.GET)
            if self.search_form.is_valid():
                self.q = self.search_form.cleaned_data["q"]
                self.is_searching = True
                images = images.search(self.q)
        else:
            self.search_form = SearchForm()

        if not self.is_searching:
            tag_name = request.GET.get("tag")
            if tag_name:
                images = images.filter(tags__name=tag_name)

        # Pagination
        paginator = Paginator(images, per_page=CHOOSER_PAGE_SIZE)
        self.images = paginator.get_page(request.GET.get("p"))
        return self.render_to_response()

    def get_context_data(self):
        return {
            "images": self.images,
            "is_searching": self.is_searching,
            "query_string": self.q,
            "will_select_format": self.request.GET.get("select_format"),
        }

    def render_to_response(self):
        raise NotImplementedError()


class ChooseView(BaseChooseView):
    def get_context_data(self):
        context = super().get_context_data()

        if permission_policy.user_has_permission(self.request.user, "add"):
            ImageForm = get_image_form(self.image_model)
            uploadform = ImageForm(
                user=self.request.user, prefix="image-chooser-upload"
            )
        else:
            uploadform = None

        collections = permission_policy.collections_user_has_permission_for(
            self.request.user, "choose"
        )
        if len(collections) < 2:
            collections = None

        context.update(
            {
                "searchform": self.search_form,
                "popular_tags": popular_tags_for_model(self.image_model),
                "collections": collections,
                "uploadform": uploadform,
            }
        )
        return context

    def render_to_response(self):
        return render_modal_workflow(
            self.request,
            "wagtailimages/chooser/chooser.html",
            None,
            self.get_context_data(),
            json_data={
                "step": "chooser",
                "error_label": _("Server Error"),
                "error_message": _(
                    "Report this error to your website administrator with the following information:"
                ),
                "tag_autocomplete_url": reverse("wagtailadmin_tag_autocomplete"),
            },
        )


class ChooseResultsView(BaseChooseView):
    def render_to_response(self):
        return TemplateResponse(
            self.request, "wagtailimages/chooser/results.html", self.get_context_data()
        )


def image_chosen(request, image_id):
    image = get_object_or_404(get_image_model(), id=image_id)

    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={"step": "image_chosen", "result": get_image_result_data(image)},
    )


def duplicate_found(request, new_image, existing_image):
    next_step_url = (
        "wagtailimages:chooser_select_format"
        if request.GET.get("select_format")
        else "wagtailimages:image_chosen"
    )
    choose_new_image_url = reverse(next_step_url, args=(new_image.id,))
    choose_existing_image_url = reverse(next_step_url, args=(existing_image.id,))

    cancel_duplicate_upload_action = (
        f"{reverse('wagtailimages:delete', args=(new_image.id,))}?"
        f"{urlencode({'next': choose_existing_image_url})}"
    )

    duplicate_upload_html = render_to_string(
        "wagtailimages/chooser/confirm_duplicate_upload.html",
        {
            "new_image": new_image,
            "existing_image": existing_image,
            "confirm_duplicate_upload_action": choose_new_image_url,
            "cancel_duplicate_upload_action": cancel_duplicate_upload_action,
        },
        request,
    )
    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={
            "step": "duplicate_found",
            "htmlFragment": duplicate_upload_html,
        },
    )


@permission_checker.require("add")
def chooser_upload(request):
    Image = get_image_model()
    ImageForm = get_image_form(Image)

    if request.method == "POST":
        image = Image(uploaded_by_user=request.user)
        form = ImageForm(
            request.POST,
            request.FILES,
            instance=image,
            user=request.user,
            prefix="image-chooser-upload",
        )

        if form.is_valid():
            # Set image file size
            image.file_size = image.file.size

            # Set image file hash
            image.file.seek(0)
            image._set_file_hash(image.file.read())
            image.file.seek(0)

            form.save()

            # Reindex the image to make sure all tags are indexed
            search_index.insert_or_update_object(image)

            duplicates = find_image_duplicates(
                image=image,
                user=request.user,
                permission_policy=permission_policy,
            )
            existing_image = duplicates.first()
            if existing_image:
                return duplicate_found(request, image, existing_image)

            if request.GET.get("select_format"):
                form = ImageInsertionForm(
                    initial={"alt_text": image.default_alt_text},
                    prefix="image-chooser-insertion",
                )
                return render_modal_workflow(
                    request,
                    "wagtailimages/chooser/select_format.html",
                    None,
                    {"image": image, "form": form},
                    json_data={"step": "select_format"},
                )
            else:
                # not specifying a format; return the image details now
                return render_modal_workflow(
                    request,
                    None,
                    None,
                    None,
                    json_data={
                        "step": "image_chosen",
                        "result": get_image_result_data(image),
                    },
                )
    else:
        form = ImageForm(user=request.user, prefix="image-chooser-upload")

    upload_form_html = render_to_string(
        "wagtailimages/chooser/upload_form.html",
        {
            "form": form,
            "will_select_format": request.GET.get("select_format"),
        },
        request,
    )

    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={"step": "reshow_upload_form", "htmlFragment": upload_form_html},
    )


def chooser_select_format(request, image_id):
    image = get_object_or_404(get_image_model(), id=image_id)

    if request.method == "POST":
        form = ImageInsertionForm(
            request.POST,
            initial={"alt_text": image.default_alt_text},
            prefix="image-chooser-insertion",
        )
        if form.is_valid():

            format = get_image_format(form.cleaned_data["format"])
            preview_image = image.get_rendition(format.filter_spec)

            image_data = {
                "id": image.id,
                "title": image.title,
                "format": format.name,
                "alt": form.cleaned_data["alt_text"],
                "class": format.classnames,
                "edit_link": reverse("wagtailimages:edit", args=(image.id,)),
                "preview": {
                    "url": preview_image.url,
                    "width": preview_image.width,
                    "height": preview_image.height,
                },
                "html": format.image_to_editor_html(
                    image, form.cleaned_data["alt_text"]
                ),
            }

            return render_modal_workflow(
                request,
                None,
                None,
                None,
                json_data={"step": "image_chosen", "result": image_data},
            )
    else:
        initial = {"alt_text": image.default_alt_text}
        initial.update(request.GET.dict())
        # If you edit an existing image, and there is no alt text, ensure that
        # "image is decorative" is ticked when you open the form
        initial["image_is_decorative"] = initial["alt_text"] == ""
        form = ImageInsertionForm(initial=initial, prefix="image-chooser-insertion")

    return render_modal_workflow(
        request,
        "wagtailimages/chooser/select_format.html",
        None,
        {"image": image, "form": form},
        json_data={"step": "select_format"},
    )



from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail.users.utils import user_can_delete_user
from wagtail.users.views.bulk_actions.user_bulk_action import UserBulkAction


class DeleteBulkAction(UserBulkAction):
    display_name = _("Delete")
    action_type = "delete"
    aria_label = _("Delete selected users")
    template_name = "wagtailusers/bulk_actions/confirm_bulk_delete.html"
    action_priority = 10
    classes = {"serious"}

    def check_perm(self, obj):
        return user_can_delete_user(self.request.user, obj)

    def get_execution_context(self):
        return {**super().get_execution_context(), "model": self.model}

    @classmethod
    def execute_action(cls, objects, model=None, **kwargs):
        if model is None:
            model = cls.get_default_model()
        model.objects.filter(pk__in=[obj.pk for obj in objects]).delete()
        return len(objects), 0

    def get_success_message(self, num_parent_objects, num_child_objects):
        return ngettext(
            "%(num_parent_objects)d user has been deleted",
            "%(num_parent_objects)d users have been deleted",
            num_parent_objects,
        ) % {"num_parent_objects": num_parent_objects}



from django.conf import settings
from django.utils.module_loading import import_string

from wagtail.admin.rich_text.editors.draftail import DraftailRichTextArea  # NOQA

DEFAULT_RICH_TEXT_EDITORS = {
    "default": {"WIDGET": "wagtail.admin.rich_text.DraftailRichTextArea"}
}


def get_rich_text_editor_widget(name="default", features=None):
    editor_settings = DEFAULT_RICH_TEXT_EDITORS.copy()
    editor_settings.update(getattr(settings, "WAGTAILADMIN_RICH_TEXT_EDITORS", {}))

    editor = editor_settings[name]
    options = editor.get("OPTIONS", None)

    if features is None and options is not None:
        # fall back on 'features' list within OPTIONS, if any
        features = options.get("features", None)

    cls = import_string(editor["WIDGET"])

    kwargs = {}

    if options is not None:
        kwargs["options"] = options

    if getattr(cls, "accepts_features", False):
        kwargs["features"] = features

    return cls(**kwargs)



from warnings import warn

from django.conf import settings

from wagtail.utils.deprecation import RemovedInWagtail50Warning


def get_admin_base_url():
    """
    Gets the base URL for the wagtail admin site. This is set in `settings.WAGTAILADMIN_BASE_URL`,
    which was previously `settings.BASE_URL`.
    """

    admin_base_url = getattr(settings, "WAGTAILADMIN_BASE_URL", None)
    if admin_base_url is None and hasattr(settings, "BASE_URL"):
        warn(
            "settings.BASE_URL has been renamed to settings.WAGTAILADMIN_BASE_URL",
            category=RemovedInWagtail50Warning,
        )
        admin_base_url = settings.BASE_URL

    return admin_base_url



# -*- coding: utf-8 -*-
import taggit.managers
from django.conf import settings
from django.db import migrations, models

import wagtail.images.models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0008_image_created_at_index"),
    ]

    operations = [
        migrations.AlterField(
            model_name="image",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="image",
            name="file",
            field=models.ImageField(
                upload_to=wagtail.images.models.get_upload_to,
                height_field="height",
                width_field="width",
                verbose_name="file",
            ),
        ),
        migrations.AlterField(
            model_name="image",
            name="height",
            field=models.IntegerField(editable=False, verbose_name="height"),
        ),
        migrations.AlterField(
            model_name="image",
            name="tags",
            field=taggit.managers.TaggableManager(
                through="taggit.TaggedItem",
                verbose_name="tags",
                blank=True,
                help_text=None,
                to="taggit.Tag",
            ),
        ),
        migrations.AlterField(
            model_name="image",
            name="title",
            field=models.CharField(max_length=255, verbose_name="title"),
        ),
        migrations.AlterField(
            model_name="image",
            name="uploaded_by_user",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                blank=True,
                null=True,
                to=settings.AUTH_USER_MODEL,
                editable=False,
                verbose_name="uploaded by user",
            ),
        ),
        migrations.AlterField(
            model_name="image",
            name="width",
            field=models.IntegerField(editable=False, verbose_name="width"),
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from wagtail.admin import messages
from wagtail.admin.action_menu import PageActionMenu
from wagtail.admin.auth import user_has_any_page_permission, user_passes_test
from wagtail.admin.side_panels import PageSidePanels
from wagtail.admin.views.pages.utils import get_valid_next_url_from_request
from wagtail.models import Page, UserPagePermissionsProxy


def revisions_index(request, page_id):
    return redirect("wagtailadmin_pages:history", page_id)


def revisions_revert(request, page_id, revision_id):
    page = get_object_or_404(Page, id=page_id).specific
    page_perms = page.permissions_for_user(request.user)
    if not page_perms.can_edit():
        raise PermissionDenied

    revision = get_object_or_404(page.revisions, id=revision_id)
    revision_page = revision.as_page_object()

    content_type = ContentType.objects.get_for_model(page)
    page_class = content_type.model_class()

    edit_handler = page_class.get_edit_handler()
    form_class = edit_handler.get_form_class()

    form = form_class(instance=revision_page)
    edit_handler = edit_handler.get_bound_panel(
        instance=revision_page, request=request, form=form
    )

    action_menu = PageActionMenu(request, view="revisions_revert", page=page)
    side_panels = PageSidePanels(
        request,
        page,
        comments_enabled=form.show_comments_toggle,
    )

    user_avatar = render_to_string(
        "wagtailadmin/shared/user_avatar.html", {"user": revision.user}
    )

    messages.warning(
        request,
        mark_safe(
            _(
                "You are viewing a previous version of this page from <b>%(created_at)s</b> by %(user)s"
            )
            % {
                "created_at": revision.created_at.strftime("%d %b %Y %H:%M"),
                "user": user_avatar,
            }
        ),
    )

    return TemplateResponse(
        request,
        "wagtailadmin/pages/edit.html",
        {
            "page": page,
            "revision": revision,
            "is_revision": True,
            "content_type": content_type,
            "edit_handler": edit_handler,
            "errors_debug": None,
            "action_menu": action_menu,
            "side_panels": side_panels,
            "preview_modes": page.preview_modes,
            "form": form,  # Used in unit tests
            "media": edit_handler.media
            + form.media
            + action_menu.media
            + side_panels.media,
        },
    )


@user_passes_test(user_has_any_page_permission)
def revisions_view(request, page_id, revision_id):
    page = get_object_or_404(Page, id=page_id).specific

    perms = page.permissions_for_user(request.user)
    if not (perms.can_publish() or perms.can_edit()):
        raise PermissionDenied

    revision = get_object_or_404(page.revisions, id=revision_id)
    revision_page = revision.as_page_object()

    try:
        preview_mode = page.default_preview_mode
    except IndexError:
        raise PermissionDenied

    return revision_page.make_preview_request(request, preview_mode)


@user_passes_test(user_has_any_page_permission)
def revisions_compare(request, page_id, revision_id_a, revision_id_b):
    page = get_object_or_404(Page, id=page_id).specific

    # Get revision to compare from
    if revision_id_a == "live":
        if not page.live:
            raise Http404

        revision_a = page
        revision_a_heading = _("Live")
    elif revision_id_a == "earliest":
        revision_a = page.revisions.order_by("created_at", "id").first()
        if revision_a:
            revision_a = revision_a.as_page_object()
            revision_a_heading = _("Earliest")
        else:
            raise Http404
    else:
        revision_a = get_object_or_404(
            page.revisions, id=revision_id_a
        ).as_page_object()
        revision_a_heading = str(
            get_object_or_404(page.revisions, id=revision_id_a).created_at
        )

    # Get revision to compare to
    if revision_id_b == "live":
        if not page.live:
            raise Http404

        revision_b = page
        revision_b_heading = _("Live")
    elif revision_id_b == "latest":
        revision_b = page.revisions.order_by("created_at", "id").last()
        if revision_b:
            revision_b = revision_b.as_page_object()
            revision_b_heading = _("Latest")
        else:
            raise Http404
    else:
        revision_b = get_object_or_404(
            page.revisions, id=revision_id_b
        ).as_page_object()
        revision_b_heading = str(
            get_object_or_404(page.revisions, id=revision_id_b).created_at
        )

    comparison = (
        page.get_edit_handler()
        .get_bound_panel(instance=page, request=request, form=None)
        .get_comparison()
    )
    comparison = [comp(revision_a, revision_b) for comp in comparison]
    comparison = [comp for comp in comparison if comp.has_changed()]

    return TemplateResponse(
        request,
        "wagtailadmin/pages/revisions/compare.html",
        {
            "page": page,
            "revision_a_heading": revision_a_heading,
            "revision_a": revision_a,
            "revision_b_heading": revision_b_heading,
            "revision_b": revision_b,
            "comparison": comparison,
        },
    )


def revisions_unschedule(request, page_id, revision_id):
    page = get_object_or_404(Page, id=page_id).specific

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_unschedule():
        raise PermissionDenied

    revision = get_object_or_404(page.revisions, id=revision_id)

    next_url = get_valid_next_url_from_request(request)

    subtitle = _('revision {0} of "{1}"').format(
        revision.id, page.get_admin_display_title()
    )

    if request.method == "POST":
        revision.approved_go_live_at = None
        revision.save(user=request.user, update_fields=["approved_go_live_at"])

        messages.success(
            request,
            _('Version {0} of "{1}" unscheduled.').format(
                revision.id, page.get_admin_display_title()
            ),
            buttons=[
                messages.button(
                    reverse("wagtailadmin_pages:edit", args=(page.id,)), _("Edit")
                )
            ],
        )

        if next_url:
            return redirect(next_url)
        return redirect("wagtailadmin_pages:history", page.id)

    return TemplateResponse(
        request,
        "wagtailadmin/pages/revisions/confirm_unschedule.html",
        {"page": page, "revision": revision, "next": next_url, "subtitle": subtitle},
    )



from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse

from wagtail.models import Page, PageLogEntry
from wagtail.test.testapp.models import SimplePage
from wagtail.test.utils import WagtailTestUtils


class TestConvertAlias(TestCase, WagtailTestUtils):
    def setUp(self):
        # Find root page
        self.root_page = Page.objects.get(id=2)

        # Add child page
        self.child_page = SimplePage(
            title="Hello world!", slug="hello-world", content="hello"
        )
        self.root_page.add_child(instance=self.child_page)

        # Add alias page
        self.alias_page = self.child_page.create_alias(update_slug="alias-page")

        # Login
        self.user = self.login()

    def test_convert_alias(self):
        response = self.client.get(
            reverse("wagtailadmin_pages:convert_alias", args=[self.alias_page.id])
        )
        self.assertEqual(response.status_code, 200)

    def test_convert_alias_not_alias(self):
        response = self.client.get(
            reverse("wagtailadmin_pages:convert_alias", args=[self.child_page.id])
        )
        self.assertEqual(response.status_code, 404)

    def test_convert_alias_bad_permission(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(
            reverse("wagtailadmin_pages:convert_alias", args=[self.alias_page.id])
        )

        # Check that the user received a permission denied response
        self.assertRedirects(response, "/admin/")

    def test_post_convert_alias(self):
        response = self.client.post(
            reverse("wagtailadmin_pages:convert_alias", args=[self.alias_page.id])
        )

        # User should be redirected to the edit view of the converted page
        self.assertRedirects(
            response, reverse("wagtailadmin_pages:edit", args=[self.alias_page.id])
        )

        # Check the page was converted
        self.alias_page.refresh_from_db()
        self.assertIsNone(self.alias_page.alias_of)

        # Check that a revision was created
        revision = self.alias_page.revisions.get()
        self.assertEqual(revision.user, self.user)
        self.assertEqual(self.alias_page.live_revision, revision)

        # Check audit log
        log = PageLogEntry.objects.get(action="wagtail.convert_alias")
        self.assertFalse(log.content_changed)
        self.assertEqual(
            log.data,
            {
                "page": {
                    "id": self.alias_page.id,
                    "title": self.alias_page.get_admin_display_title(),
                }
            },
        )
        self.assertEqual(log.page, self.alias_page.page_ptr)
        self.assertEqual(log.revision, revision)
        self.assertEqual(log.user, self.user)



# Generated by Django 4.0.2 on 2022-02-28 01:28

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0067_alter_pagerevision_content_json"),
        ("contenttypes", "0002_remove_content_type_name"),
        ("tests", "0065_alter_extendedformfield_choices_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="PageWithGenericRelation",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wagtailcore.page",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="RelatedGenericRelation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveBigIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailforms", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="formsubmission",
            options={"verbose_name": "Form Submission"},
        ),
        migrations.AlterField(
            model_name="formsubmission",
            name="submit_time",
            field=models.DateTimeField(auto_now_add=True, verbose_name="Submit time"),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-22 09:34
from django.db import migrations, models

import wagtail.models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0023_alter_page_revision_on_delete_behaviour"),
    ]

    operations = [
        migrations.AlterField(
            model_name="page",
            name="content_type",
            field=models.ForeignKey(
                on_delete=models.SET(wagtail.models.get_default_page_content_type),
                related_name="pages",
                to="contenttypes.ContentType",
                verbose_name="content type",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations


def initial_data(apps, schema_editor):
    ContentType = apps.get_model("contenttypes.ContentType")
    Group = apps.get_model("auth.Group")
    Page = apps.get_model("wagtailcore.Page")
    Site = apps.get_model("wagtailcore.Site")
    GroupPagePermission = apps.get_model("wagtailcore.GroupPagePermission")

    # Create page content type
    page_content_type, created = ContentType.objects.get_or_create(
        model="page", app_label="wagtailcore"
    )

    # Create root page
    root = Page.objects.create(
        title="Root",
        slug="root",
        content_type=page_content_type,
        path="0001",
        depth=1,
        numchild=1,
        url_path="/",
    )

    # Create homepage
    homepage = Page.objects.create(
        title="Welcome to your new Wagtail site!",
        slug="home",
        content_type=page_content_type,
        path="00010001",
        depth=2,
        numchild=0,
        url_path="/home/",
    )

    # Create default site
    Site.objects.create(
        hostname="localhost", root_page_id=homepage.id, is_default_site=True
    )

    # Create auth groups
    moderators_group = Group.objects.create(name="Moderators")
    editors_group = Group.objects.create(name="Editors")

    # Create group permissions
    GroupPagePermission.objects.create(
        group=moderators_group,
        page=root,
        permission_type="add",
    )
    GroupPagePermission.objects.create(
        group=moderators_group,
        page=root,
        permission_type="edit",
    )
    GroupPagePermission.objects.create(
        group=moderators_group,
        page=root,
        permission_type="publish",
    )

    GroupPagePermission.objects.create(
        group=editors_group,
        page=root,
        permission_type="add",
    )
    GroupPagePermission.objects.create(
        group=editors_group,
        page=root,
        permission_type="edit",
    )


def remove_initial_data(apps, schema_editor):
    """This function does nothing. The below code is commented out together
    with an explanation of why we don't need to bother reversing any of the
    initial data"""
    pass
    # This does not need to be deleted, Django takes care of it.
    # page_content_type = ContentType.objects.get(
    #     model='page',
    #     app_label='wagtailcore',
    # )

    # Page objects: Do nothing, the table will be deleted when reversing 0001

    # Do not reverse Site creation since other models might depend on it

    # Remove auth groups -- is this safe? External objects might depend
    # on these groups... seems unsafe.
    # Group.objects.filter(
    #     name__in=('Moderators', 'Editors')
    # ).delete()
    #
    # Likewise, we're leaving all GroupPagePermission unchanged as users may
    # have been assigned such permissions and its harmless to leave them.


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(initial_data, remove_initial_data),
    ]
